from copy import copy
import threading
import functools
import operator
import re

from contextlib import contextmanager
from visidata import VisiData, Sheet, IndexSheet, vd, date, anytype, vlen, clipdraw, colors, stacktrace, PyobjSheet, BaseSheet, ExpectedException
from visidata import ItemColumn, AttrColumn, Column, TextSheet, asyncthread, wrapply, ColumnsSheet, UNLOADED, ExprColumn, undoAttrCopyFunc, ENTER

vd.option('disp_ibis_sidebar', 'pending_sql', 'which sidebar property to display')
vd.option('sql_always_count', False, 'whether to include count of total number of results')
vd.option('ibis_limit', 500, 'max number of rows to get in query')


def vdtype_to_ibis_type(t):
    from ibis.expr import datatypes as dt
    return {
        int: dt.int,
        float: dt.float,
        date: dt.date,
        str: dt.string,
    }.get(t)


def dtype_to_vdtype(dtype):
    from ibis.expr import datatypes as dt

    try:
        if isinstance(dtype, dt.Decimal):
            if dtype.scale == 0:
                return int
            else:
                return float
        if isinstance(dtype, dt.Integer):
            return int
        if isinstance(dtype, dt.Floating):
            return float
        if isinstance(dtype, (dt.Date, dt.Timestamp)):
            return date
    except TypeError:
        # For categoricals and other pandas-defined dtypes
        pass
    return anytype


@VisiData.api
def configure_ibis(vd):
    import ibis

    vd.aggregator('collect', ibis.expr.types.AnyValue.collect, 'collect a list of values')
    ibis.options.verbose_log = vd.status
    if vd.options.debug:
        ibis.options.verbose = True

    if 'ibis_type' not in set(c.expr for c in ColumnsSheet.columns):
        ColumnsSheet.columns += [
            AttrColumn('ibis_type', type=str)
        ]


@VisiData.api
def open_vdsql(vd, p, filetype=None):
    import ibis

    vd.configure_ibis()

    # on-demand aliasing, so we don't need deps for all backends
    ext_aliases = dict(db='sqlite', ddb='duckdb', sqlite3='sqlite')
    if p.ext in ext_aliases:
        setattr(ibis, p.ext, ext_aliases.get(p.ext))

    return IbisTableIndexSheet(p.base_stem, source=p, filetype=None, database_name=None,
                               ibis_conpool=IbisConnectionPool(p), sheet_type=IbisTableSheet)


vd.open_ibis = vd.open_vdsql
vd.openurl_sqlite = vd.open_vdsql


class IbisConnectionPool:
    def __init__(self, source, pool=None, total=0):
        self.source = source
        self._local = threading.local()
        self._local.connection = None

    def __copy__(self):
        return IbisConnectionPool(self.source)

    @contextmanager
    def get_conn(self):
        import ibis
        if not hasattr(self._local, 'connection') or not self._local.connection:
            self._local.connection = ibis.connect(str(self.source))
        yield self._local.connection


class IbisTableIndexSheet(IndexSheet):
    # sheet_type = IbisTableSheet  # set below

    @property
    def con(self):
        return self.ibis_conpool.get_conn()

    def rawSql(self, qstr):
        with self.con as con:
            return IbisTableSheet('rawsql',
                                  ibis_conpool=self.ibis_conpool,
                                  ibis_source=qstr,
                                  source=qstr,
                                  query=con.sql(qstr))

    def iterload(self):
        with self.con as con:
            if self.database_name:
                con.set_database(self.database_name)

            # use the actual count instead of the returned limit
            nrows_col = self.column('rows')
            nrows_col.expr = 'countRows'
            nrows_col.width += 3

            for tblname in con.list_tables():
                yield self.sheet_type(tblname,
                        ibis_source=self.source,
                        ibis_filetype=self.filetype,
                        ibis_conpool=self.ibis_conpool,
                        database_name=self.database_name,
                        table_name=tblname,
                        source=self.source,
                        query=None)


class IbisColumn(ItemColumn):
    @property
    def ibis_type(self):
        return self.sheet.query[self.ibis_name].type()

    @asyncthread
    def memo_aggregate(self, agg, rows):
        'Show aggregated value in status, and add ibis expr to memory for use later.'

        aggexpr = self.ibis_aggr(agg.name)  # ignore rows, do over whole query

        with self.sheet.con as con:
            aggval = con.execute(aggexpr)

        typedval = wrapply(agg.type or self.type, aggval)
        dispval = self.format(typedval)
        k = self.name+'_'+agg.name
        vd.status(f'{k}={dispval}')
        vd.memory[k] = aggval
        # store aggexpr somewhere to use in later subquery

    def expand(self, rows):
        return self.expand_struct(rows)

    def expand_struct(self, rows):
        oldexpr = self.sheet.ibis_current_expr
        struct_field = self.get_ibis_col(oldexpr)
#        if struct_field is not StructType:
#            vd.fail('vdsql can only expand Struct columns')

        struct_fields = [struct_field[name] for name in struct_field.names]
        expandedCols = super().expand(rows)   # this must go after ibis_current_expr, because it alters ibis_current_expr
        fields = []
        for ibiscol, expcol in zip(struct_fields, expandedCols):
            fields.append(ibiscol.name(expcol.name))
    #    self.sheet.query = oldexpr.drop([struct_field.get_name()]).mutate(fields)
        self.sheet.query = oldexpr.mutate(fields)
        return expandedCols


class LazyIbisColMap:
    def __init__(self, sheet, q):
        self._sheet = sheet
        self._query = q
        self._map = {col.name: col for col in sheet.columns}

    def __getitem__(self, k):
        col = self._map[k]
        return col.get_ibis_col(self._query)


class IbisTableSheet(Sheet):
    @property
    def con(self):
        return self.ibis_conpool.get_conn()

    def choose_sidebar(self):
        sidebars = ['base_sql', 'pending_sql', 'ibis_current_expr', 'curcol_sql', 'pending_expr']
        opts = []
        for s in sidebars:
            try:
                opts.append({'key': s, 'value':getattr(self, s)})
            except Exception as e:
                if self.options.debug:
                    vd.exceptionCaught()

        vd.options.disp_ibis_sidebar = vd.chooseOne(opts)

    @property
    def curcol_sql(self):
        expr = self.cursorCol.get_ibis_col(self.ibis_current_expr)
        if expr is not None:
            return self.ibis_expr_to_sql(expr, fragment=True)

    def ibis_expr_to_sql(self, expr, fragment=False):
        import sqlparse
        with self.con as con:
            context = con.compiler.make_context()
            trclass = con.compiler.translator_class(expr.op(), context=context)
            if fragment:
                compiled = trclass.get_result()
            else:
                compiled = con.compile(expr)
            if not isinstance(compiled, str):
                compiled = str(compiled.compile(compile_kwargs={'literal_binds': True}))
        return sqlparse.format(compiled, reindent=True, keyword_case='upper', wrap_after=40)

    @property
    def sidebar(self) -> str:
        sbtype = self.options.disp_ibis_sidebar
        if sbtype:
            txt = str(getattr(self, sbtype, ''))
            if txt:
                return f'# {sbtype}\n'+txt

    @property
    def ibis_locals(self):
        return LazyIbisColMap(self, self.query)

    def select_row(self, row):
        k = self.rowkey(row) or vd.fail('need key column to select individual rows')
        super().selectRow(row)
        self.ibis_selection.append(self.matchRowKeyExpr(row))

    def stoggle_row(self, row):
        vd.fail('cannot toggle selection of individual row in vdsql')

    def unselect_row(self, row):
        super().unselectRow(row)
        self.ibis_selection = [ self.ibis_filter & ~self.matchRowKeyExpr(row) ]

    def matchRowKeyExpr(self, row):
        import ibis
        k = self.rowkey(row) or vd.fail('need key column to select individual rows')

        return functools.reduce(operator.and_, [
            c.get_ibis_col(self.query, typed=True) == k[i]
                for i, c in enumerate(self.keyCols)
        ])

    @property
    def ibis_current_expr(self):
        return self.get_current_expr(typed=False)

    def get_current_expr(self, typed=False):
        q = self.query
        extra_cols = {}
        for c in self.visibleCols:
            ibis_col = c.get_ibis_col(q, typed=typed)
            if ibis_col is not None:
                extra_cols[c.name] = ibis_col
            else:
                vd.warning(f'no column {c.name}')

        if extra_cols:
            q = q.mutate(**extra_cols)

        return q

    @property
    def ibis_filter(self):
        import ibis
        selectors = [self.ibisCompileExpr(f, self.get_current_expr(typed=True)) for f in self.ibis_selection]
        if not selectors:
            return ibis.literal(True)
        return functools.reduce(operator.or_, selectors)

    @property
    def pending_expr(self):
        import ibis
        q = self.get_current_expr(typed=True)
        if self.ibis_selection:
            q = q.filter(self.ibis_filter)

        if self._ordering:
            colorder = []
            for col, rev in self._ordering:
                ibiscol = col.get_ibis_col(q)  #1856
                if rev:
                    ibiscol = ibis.desc(ibiscol)
                colorder.append(ibiscol)
            q = q.order_by(colorder)

        return q

    def ibisCompileExpr(self, expr, q):
        if isinstance(expr, str):
            return eval(expr, vd.getGlobals(), LazyIbisColMap(self, q))
        else:
            return expr

    def evalIbisExpr(self, expr):
        return eval(expr, vd.getGlobals(), self.ibis_locals)

    @property
    def base_sql(self):
        return self.sqlize(self.ibis_current_expr)

    @property
    def pending_sql(self):
        return self.sqlize(self.pending_expr)

    def sqlize(self, expr):
        if vd.options.debug:
            expr = self.withRowcount(expr)
        return self.ibis_expr_to_sql(expr)

    @property
    def substrait(self):
        from ibis_substrait.compiler.core import SubstraitCompiler
        compiler = SubstraitCompiler()
        return compiler.compile(self.ibis_current_expr)

    def withRowcount(self, q):
        if self.options.sql_always_count:
            # return q.mutate(__n__=q.count())
            return q.cross_join(q.aggregate(__n__=lambda t: t.count()))
        return q

    def beforeLoad(self):
        self.options.disp_rstatus_fmt = self.options.disp_rstatus_fmt.replace('nRows', 'countRows')
        self.options.disp_rstatus_fmt = self.options.disp_rstatus_fmt.replace('nSelectedRows', 'countSelectedRows')

    def baseQuery(self, con):
        'Return base table for {database_name}.{table_name}'
        import ibis
        tbl = con.table(self.table_name)
        return ibis.table(tbl.schema(), name=self.fqtblname(con))

    def fqtblname(self, con) -> str:
        'Return fully-qualified table name including database/schema, or whatever connection needs to identify this table.'
        if hasattr(con, '_fully_qualified_name'):
            return con._fully_qualified_name(self.table_name, self.database_name)
        return self.table_name

    def iterload(self):
        with self.con as con:
            if self.query is None:
                self.query = self.baseQuery(con)

            self.reloadColumns(self.query)  # columns based on query without metadata
            self.query_result = con.execute(self.withRowcount(self.query),
                                            limit=self.options.ibis_limit or None)

            yield from self.query_result.itertuples()


    def reloadColumns(self, expr, start=1):
        oldkeycols = {c.name:c for c in self.keyCols}
        self._nrows_col = -1

        for i, (colname, dtype) in enumerate(expr.schema().items(), start=start):
            keycol=oldkeycols.get(colname, Column()).keycol
            if i-start < self.nKeys:
                keycol = i+1

            if colname == '__n__':
                self._nrows_col = i
                continue

            self.addColumn(IbisColumn(colname, i,
                           type=dtype_to_vdtype(dtype),
                           keycol=keycol,
                           ibis_name=colname))

    @property
    def countSelectedRows(self):
        return f'{self.nSelectedRows}+'

    @property
    def countRows(self):
        if self.rows is UNLOADED:
            return None
        if not self.rows or self._nrows_col < 0:
            return self.nRows
        return self.rows[0][self._nrows_col]  # __n__

    def groupBy(self, groupByCols):
        from ibis import _
        import ibis
        from ibis.expr import datatypes as dt
        aggr_cols = [groupByCols[0].ibis_col.count().name('count')]
        for c in self.visibleCols:
            aggr_cols.extend(c.ibis_aggrs)

        q = self.ibis_current_expr
        groupq = q.aggregate(aggr_cols, by=[c.ibis_col for c in groupByCols])
        try:
            win = ibis.window(order_by=ibis.NA)
        except ibis.common.exceptions.IbisTypeError: # ibis bug: there is not yet a good workaround that covers all backends
            win = ibis.window(order_by=None)
        groupq = groupq.mutate(percent=_['count']*100 / _['count'].sum().over(win))

        histolen = 40
        histogram_char = self.options.disp_histogram
        if histogram_char and len(aggr_cols) == 1:
            groupq = groupq.mutate(maxcount=_['count'].max())
            hval = ibis.literal(histogram_char, type=dt.string)

            def _histogram(t):
                return hval.repeat((histolen*t['count']/t.maxcount).cast(dt.int))

            groupq = groupq.mutate(histogram=_histogram)

        groupq = groupq.order_by(ibis.desc('count'))

        return IbisFreqTable(self.name, *(col.name for col in groupByCols), 'freq',
                             ibis_conpool=self.ibis_conpool,
                             ibis_source=self.ibis_source,
                             source=self,
                             groupByCols=groupByCols,
                             query=groupq,
                             nKeys=len(groupByCols))

    def unfurl_col(self, col):
        vs = copy(self)
        vs.names = [self.name, col.name, 'unfurled']
        vs.query = self.ibis_current_expr.mutate(**{col.name:col.ibis_col.unnest()})
        vs.cursorVisibleColIndex = self.cursorVisibleColIndex
        return vs

    def openJoin(self, others, jointype=''):
        sheets = [self] + others

        sheets[1:] or vd.fail("join requires more than 1 sheet")

        if jointype == 'append':
            q = self.ibis_current_expr
            for other in others:
                q = q.union(other.ibis_current_expr)
            return IbisTableSheet('&'.join(vs.name for vs in sheets), query=q, ibis_source=self.ibis_source, ibis_conpool=self.ibis_conpool)

        for s in sheets:
            s.keyCols or vd.fail(f'{s.name} has no key cols to join')

        if jointype in ['extend', 'outer']:
            jointype = 'left'
        elif jointype in ['full']:
            jointype = 'outer'
#        elif jointype in ['inner']:
#            jointype = 'inner'


        q = self.ibis_current_expr
        for other in others:
            preds = [(a.ibis_col == b.ibis_col) for a, b in zip(self.keyCols, other.keyCols)]
            q = q.join(other.ibis_current_expr, predicates=preds, how=jointype, suffixes=('', '_'+other.name))

        return IbisTableSheet('+'.join(vs.name for vs in sheets), sources=sheets, query=q, ibis_source=self.ibis_source, ibis_conpool=self.ibis_conpool)


@Column.property
def ibis_col(col):
    return col.get_ibis_col(col.sheet.ibis_current_expr)


@Column.api
def get_ibis_col(col, query:'ibis.Expr', typed=False) -> 'ibis.Expr':
    'Return ibis.Expr for `col` within context of `query`, cast by VisiData column type if `typed`.'
    import ibis.common.exceptions

    r = None
    if isinstance(col, ExprColumn):
        r = col.sheet.evalIbisExpr(col.expr)
    elif isinstance(col, vd.ExpandedColumn):
        r = query[col.name]
    elif not hasattr(col, 'ibis_name'):
        return
    else:
        try:
            r = query[col.ibis_name]
        except (ibis.common.exceptions.IbisTypeError, AttributeError):
            r = query[col.name]

    if r is None:
        return r

    if typed:
        import ibis.expr.datatypes as dt
        if col.type is str: r = r.cast(dt.string)
        if col.type is int: r = r.cast(dt.int)
        if col.type is float: r = r.cast(dt.float)
        if col.type is date:
            if not isinstance(r.type(), (dt.Timestamp, dt.Date)):
                r = r.cast(dt.date)

    r = r.name(col.name)
    return r


@Column.property
def ibis_aggrs(col):
    return [col.ibis_aggr(aggname) for aggname in (col.aggstr or '').split()]


@Column.api
def ibis_aggr(col, aggname):
    aggname = {
        'avg': 'mean',
        'median': 'approx_median',
        'mode': 'notimpl',
        'distinct': 'nunique',
        'list': 'collect',
        'stdev': 'std',
#        'p99': 'quantile(0.99)',
#        'q10': 'quantile([.1,.2,.3,.4,.5,.6,.7,.8,.9])',
    }.get(aggname, aggname)

    agg = getattr(col.ibis_col, aggname)
    return agg().name(f'{aggname}_{col.name}')


IbisTableSheet.init('ibis_selection', list, copy=False)
IbisTableSheet.init('_sqlscr', lambda: None, copy=False)
IbisTableSheet.init('query_result', lambda: None, copy=False)
IbisTableSheet.init('ibis_conpool', lambda: None, copy=True)

@IbisTableSheet.api
def stoggle_rows(sheet):
    sheet.toggle(sheet.rows)
    sheet.ibis_selection = [~sheet.ibis_filter]


@IbisTableSheet.api
def clearSelected(sheet):
    super(IbisTableSheet, sheet).clearSelected()
    sheet.ibis_selection.clear()



@IbisTableSheet.api
def addUndoSelection(sheet):
    super(IbisTableSheet, sheet).addUndoSelection()
    vd.addUndo(undoAttrCopyFunc([sheet], 'ibis_selection'))


@IbisTableSheet.api
def select_equal_cell(sheet, col, typedval):
    if sheet.isNullFunc()(typedval):
        expr = col.ibis_col.isnull()
    else:
        q = sheet.get_current_expr(typed=True)
        ibis_col = col.get_ibis_col(q, typed=True)
        expr = (ibis_col == typedval)

    sheet.ibis_selection.append(expr)
    sheet.select(sheet.gatherBy(lambda r,c=col,v=typedval: c.getTypedValue(r) == v), progress=False)


@IbisTableSheet.api
def select_col_regex(sheet, col, regex):
    sheet.selectByIdx(vd.searchRegex(sheet, regex=regex, columns="cursorCol"))
    sheet.ibis_selection.append(col.get_ibis_col(col.sheet.query).re_search(regex))


@IbisTableSheet.api
def select_expr(sheet, expr):
    sheet.select(sheet.gatherBy(lambda r, sheet=sheet, expr=expr: sheet.evalExpr(expr, r)), progress=False)
    sheet.ibis_selection.append(expr)


@IbisTableSheet.api
def addcol_split(sheet, col, delim):
    from ibis.expr import datatypes as dt
    c = Column(col.name+'_split',
               getter=lambda col,row: col.origCol.getDisplayValue(row).split(col.expr),
               expr=delim,
               origCol=col,
               ibis_name=col.name+'_split')
    sheet.query = sheet.query.mutate(**{c.name:col.get_ibis_col(sheet.query).cast(dt.string).split(delim)})
    return c


@IbisTableSheet.api
def addcol_subst(sheet, col, before='', after=''):
    c = Column(col.name + "_re",
               getter=lambda col,row,before=before,after=after: re.sub(before, after, col.origCol.getDisplayValue(row)),
               origCol=col,
               ibis_name=col.name + "_re")
    sheet.query = sheet.query.mutate(**{c.name:col.get_ibis_col(sheet.query).re_replace(before, after)})
    return c


@IbisTableSheet.api
def addcol_cast(sheet, col):
    # sheet.query and sheet.ibis_current_expr don't match
    new_type = vdtype_to_ibis_type(col.type)
    if new_type is None:
        vd.warning(f"no type for vd type {col.type}")
        return
    expr = sheet.query[col.name].cast(new_type)
    sheet.query = sheet.query.mutate(**{col.name: expr})
    newcol = copy(col)
    col.hide()
    sheet.addColumnAtCursor(newcol)

# disable not implemented commands

@BaseSheet.api
def notimpl(sheet):
    vd.fail(f"{vd.activeCommand.longname} not implemented for {type(sheet).__name__}; copy to new non-ibis sheet with g'")


dml_cmds = '''addcol-bulk addcol-new add-row add-rows
copy-cell copy-cells copy-row copy-selected commit-sheet cut-cell cut-cells cut-row cut-selected delete-cell delete-cells delete-row delete-selected
edit-cell paste-after paste-before paste-cell setcell-expr
setcol-clipboard setcol-expr setcol-fake setcol-fill setcol-format-enum setcol-formatter setcol-incr setcol-incr-step setcol-input setcol-iter setcol-subst setcol-subst-all
'''.split()

neverimpl_cmds = '''
select-after select-around-n select-before select-equal-row select-error stoggle-after stoggle-before stoggle-row unselect-after unselect-before select-cols-regex unselect-cols-regex transpose
'''.split()

notimpl_cmds = '''
addcol-capture addcol-incr addcol-incr-step addcol-window
contract-col expand-col-depth expand-cols expand-cols-depth melt melt-regex pivot random-rows
select-error-col select-exact-cell select-exact-row select-rows
describe-sheet freq-summary
cache-col cache-cols
dive-selected-cells
dup-rows dup-rows-deep dup-selected-deep
'''.split()

for longname in list(notimpl_cmds) + list(neverimpl_cmds) + list(dml_cmds):
    if longname:
        IbisTableSheet.addCommand('', longname, 'notimpl()')

@IbisTableSheet.api
def dup_selected(sheet):
    vs=copy(sheet)
    vs.query=sheet.pending_expr
    vs.incrementName()
    vd.push(vs)


@BaseSheet.api
def incrementName(sheet):
    if isinstance(sheet.names[-1], int):
        sheet.names[-1] += 1
    else:
        sheet.names = list(sheet.names) + [1]
    sheet.name = '_'.join(map(str, sheet.names))


@IbisTableSheet.api
def dup_limit(sheet, limit:int):
    vs=copy(sheet)
    vs.name += f"_top{limit}" if limit else "_all"
    vs.query=sheet.pending_expr
    vs.options.ibis_limit=limit
    return vs

@IbisTableSheet.api
def rawSql(sheet, qstr):
    with sheet.con as con:
        return IbisTableSheet('rawsql',
                          ibis_conpool=sheet.ibis_conpool,
                          ibis_source=qstr,
                          source=qstr,
                          query=con.sql(qstr))

class IbisFreqTable(IbisTableSheet):
    def freqExpr(self, row):
        # matching key of grouped columns
        return functools.reduce(operator.and_, [
            c.get_ibis_col(self.source.query, typed=True) == self.rowkey(row)[i]
                for i, c in enumerate(self.groupByCols)
        ])

    def selectRow(self, row):
        super().selectRow(row)
        self.source.select(self.gatherBy(lambda r, sheet=self, expr=self.freqExpr(row): sheet.evalExpr(expr, r)), progress=False)
        self.source.ibis_selection.append(self.freqExpr(row))

    def openRow(self, row):
        vs = copy(self.source)
        vs.names = list(vs.names) + ['_'.join(str(x) for x in self.rowkey(row))]
        vs.query = self.source.query.filter(self.freqExpr(row))
        return vs

    def openRows(self, rows):
        'Return sheet with union of all selected items.'
        vs = copy(self.source)
        vs.names = list(vs.names) + ['several']

        vs.query = self.source.query.filter([
            functools.reduce(operator.or_, [self.freqExpr(row) for row in rows])
        ])
        return vs


IbisTableSheet.addCommand('F', 'freq-col', 'vd.push(groupBy([cursorCol]))')
IbisTableSheet.addCommand('gF', 'freq-keys', 'vd.push(groupBy(keyCols))')

IbisTableSheet.addCommand('"', 'dup-selected', 'vd.push(dup_selected())', 'open duplicate sheet with selected rows (default limit)')
IbisTableSheet.addCommand('z"', 'dup-limit', 'vd.push(dup_limit(input("max rows: ", value=options.ibis_limit)))', 'open duplicate sheet with only selected rows (input limit)')
IbisTableSheet.addCommand('gz"', 'dup-nolimit', 'vd.push(dup_limit(0))', 'open duplicate sheet with only selected rows (no limit--be careful!)')

IbisTableSheet.addCommand("'", 'addcol-cast', 'addcol_cast(cursorCol)')

IbisTableSheet.addCommand('zb', 'sidebar-choose', 'choose_sidebar()', 'choose vdsql sidebar to show')
IbisTableSheet.addCommand('', 'exec-sql', 'vd.push(rawSql(input("SQL query: ")))', 'open sheet with results of raw SQL query')
IbisTableSheet.addCommand('', 'addcol-subst', 'addColumnAtCursor(addcol_subst(cursorCol, **inputRegexSubstOld("transform column by regex: ")))') # deprecated
IbisTableSheet.addCommand('', 'addcol-regex-subst', 'addColumnAtCursor(addcol_subst(cursorCol, **inputRegexSubst("transform column by regex: ")))')
IbisTableSheet.addCommand('', 'addcol-split', 'addColumnAtCursor(addcol_split(cursorCol, input("split by delimiter: ", type="delim-split")))')
IbisTableSheet.addCommand('gt', 'stoggle-rows', 'stoggle_rows()', 'select rows matching current cell in current column')
IbisTableSheet.addCommand(',', 'select-equal-cell', 'select_equal_cell(cursorCol, cursorTypedValue)', 'select rows matching current cell in current column')
IbisTableSheet.addCommand('t', 'stoggle-row', 'stoggle_row(cursorRow); cursorDown(1)', 'toggle selection of current row')
IbisTableSheet.addCommand('s', 'select-row', 'select_row(cursorRow); cursorDown(1)', 'select current row')
IbisTableSheet.addCommand('u', 'unselect-row', 'unselect_row(cursorRow); cursorDown(1)', 'unselect current row')
#IbisTableSheet.addCommand('g,', 'select-equal-row', 'select(gatherBy(lambda r,currow=cursorRow,vcols=visibleCols: all([c.getDisplayValue(r) == c.getDisplayValue(currow) for c in vcols])), progress=False)', 'select rows matching current row in all visible columns')
#IbisTableSheet.addCommand('z,', 'select-exact-cell', 'select(gatherBy(lambda r,c=cursorCol,v=cursorTypedValue: c.getTypedValue(r) == v), progress=False)', 'select rows matching current cell in current column')
#IbisTableSheet.addCommand('gz,', 'select-exact-row', 'select(gatherBy(lambda r,currow=cursorRow,vcols=visibleCols: all([c.getTypedValue(r) == c.getTypedValue(currow) for c in vcols])), progress=False)', 'select rows matching current row in all visible columns')

IbisTableSheet.addCommand('', 'select-col-regex', 'select_col_regex(cursorCol, inputRegex("select regex: ", type="regex", defaultLast=True))', 'select rows matching regex in current column')

IbisTableSheet.addCommand('z|', 'select-expr', 'expr=inputExpr("select by expr: "); select_expr(expr)', 'select rows matching Python expression in any visible column')
IbisTableSheet.addCommand('z\\', 'unselect-expr', 'expr=inputExpr("unselect by expr: "); unselect(gatherBy(lambda r, sheet=sheet, expr=expr: sheet.evalExpr(expr, r)), progress=False)', 'unselect rows matching Python expression in any visible column')

IbisFreqTable.addCommand('g'+ENTER, 'open-selected', 'vd.push(openRows(selectedRows))')
IbisTableIndexSheet.addCommand('', 'exec-sql', 'vd.push(rawSql(input("SQL query: ")))', 'open sheet with results of raw SQL query')

IbisTableIndexSheet.class_options.load_lazy = True
IbisTableIndexSheet.sheet_type = IbisTableSheet
IbisTableSheet.class_options.clean_names = True
IbisTableSheet.class_options.regex_flags = ''

vd.addMenuItem('View', 'Sidebar', 'choose', 'sidebar-choose')



import collections

from visidata import globalCommand, BaseSheet, Column, options, vd, anytype, ENTER, asyncthread, Sheet, IndexSheet
from visidata import CellColorizer, RowColorizer, JsonLinesSheet, AttrDict
from visidata import ColumnAttr, ItemColumn
from visidata import TsvSheet, Path, Option
from visidata import undoAttrFunc, VisiData, vlen

vd.option('visibility', 0, 'visibility level (0=low, 1=high)')

vd_system_sep = '\t'


class ColumnsSheet(Sheet):
    rowtype = 'columns'
    _rowtype = Column
    _coltype = ColumnAttr
    precious = False
    guide = '''# Columns Sheet
This is a list of {sheet.nSourceCols} columns on {sheet.displaySource}.
Edit values on this sheet to change the appearance of the source sheet.
For example, edit the _{sheet.cursorCol.name}_ column to **{sheet.cursorCol.formatted_help}**.

Some new commands on this sheet operate on all selected columns on the source sheet:

- {help.commands.hide_selected}
- {help.commands.key_selected}
- {help.commands.key_off_selected}
- {help.commands.type_int_selected}
- or `g` with any standard type to set type of selected source columns to that type

Other commands (not specific to Columns Sheet):

- {help.commands.setcol_input}
'''

    class ValueColumn(Column):
        'passthrough to the value on the source cursorRow'
        def calcValue(self, srcCol):
            return srcCol.getDisplayValue(srcCol.sheet.cursorRow)
        def setValue(self, srcCol, val, setModified=True):
            srcCol.setValue(srcCol.sheet.cursorRow, val, setModified=setModified)

    columns = [
            ColumnAttr('sheet', type=str),
            ColumnAttr('name', help='rename the column on the source sheet'),
            ColumnAttr('keycol', type=int, width=0),
            ColumnAttr('width', type=int, help='set the column width (`0` to hide completely)'),
            ColumnAttr('height', type=int, disp_expert=1, help='set a maximum height for the row, if this column will fill it'),
            ColumnAttr('hoffset', type=int, width=0),
            ColumnAttr('voffset', type=int, width=0),
            ColumnAttr('type', 'typestr', help='convert all values to a specific type'),
            ColumnAttr('fmtstr', help='use a custom format string, either C-style (`%0.4f`) or Python-style (`{{:0.4f}}`)'),
            ColumnAttr('formatter', disp_expert=1, help='use a custom format function (**{col.help_formatters}**)'),
            ColumnAttr('displayer', disp_expert=1, help='use a custom display function (**{col.help_displayers}**)'),
            ValueColumn('value', help='change the value of this cell on the source sheet'),
            ColumnAttr('expr', disp_expert=1, help='change the main column parameter'),
            ColumnAttr('ncalcs', type=int, width=0, cache=False),
            ColumnAttr('maxtime', type=float, width=0, cache=False),
            ColumnAttr('totaltime', type=float, width=0, cache=False),
    ]
    nKeys = 2
    colorizers = [
        RowColorizer(7, 'color_key_col', lambda s,c,r,v: r and r.keycol),
        RowColorizer(8, 'color_hidden_col', lambda s,c,r,v: r and r.hidden),
    ]

    @property
    def nSourceCols(self):
        return sum(vs.nCols for vs in self.source)

    def loader(self):
        if len(self.source) == 1:
            self.rows = self.source[0].columns
            self.cursorRowIndex = self.source[0].cursorColIndex
            self.columns[0].hide()  # hide 'sheet' column if only one sheet
        else:
            self.rows = [col for vs in self.source for col in vs.visibleCols if isinstance(vs, Sheet) and vs is not self]

    def newRow(self):
        c = type(self.source[0])._coltype()
        c.recalc(self.source[0])
        return c

class MetaSheet(Sheet):
    pass

class VisiDataMetaSheet(TsvSheet):
    pass

# commandline must not override these for internal sheets
VisiDataMetaSheet.options.delimiter = vd_system_sep
VisiDataMetaSheet.options.header = 1
VisiDataMetaSheet.options.skip = 0
VisiDataMetaSheet.options.row_delimiter = '\n'
VisiDataMetaSheet.options.encoding = 'utf-8'


@VisiData.property
def allColumnsSheet(vd):
    return ColumnsSheet("all_columns", source=vd.stackedSheets)


@VisiData.api
def save_visidatarc(vd, p, vs):
    with p.open(mode='w') as fp:
        for opt in vs.rows:
            rval = repr(opt.value)
            defopt = vd.options._get(opt.name, 'default')
            leading = '# ' if opt.value == defopt.value else ''
            fp.write(f'{leading}options.{opt.name:25s} = {rval:10s}  # {defopt.helpstr}\n')


ColumnsSheet.addCommand('', 'join-cols', 'sheet.join_cols()', 'add column from concatenating selected source columns')

@ColumnsSheet.api
def join_cols(sheet):
    cols = sheet.onlySelectedRows
    destSheet = cols[0].sheet

    if len(set(c.sheet for c in cols)) > 1:
        vd.fail('joined columns must come from the same source sheet')

    c = Column(options.name_joiner.join(c.name for c in cols),
                getter=lambda col,row,cols=cols,ch=options.value_joiner: ch.join(c.getDisplayValue(row) for c in cols))

    vd.status(f"added {c.name} to {destSheet}")
    destSheet.addColumn(c, index=sheet.cursorRowIndex)


# copy vd.sheets so that ColumnsSheet itself isn't included (for recalc in addRow)
globalCommand('gC', 'columns-all', 'vd.push(vd.allColumnsSheet)', 'open Columns Sheet: edit column properties for all visible columns from all sheets on the sheets stack')

Sheet.addCommand('C', 'columns-sheet', 'vd.push(ColumnsSheet(name+"_columns", source=[sheet]))', 'open Columns Sheet: edit column properties for current sheet')

# used ColumnsSheet, affecting the 'row' (source column)
ColumnsSheet.addCommand('g!', 'key-selected', 'for c in onlySelectedRows: c.sheet.setKeys([c])', 'set selected source columns as key columns')
ColumnsSheet.addCommand('gz!', 'key-off-selected', 'for c in onlySelectedRows: c.sheet.unsetKeys([c])', 'unset selected source columns as key columns')

ColumnsSheet.addCommand('g-', 'hide-selected', 'onlySelectedRows.hide()', 'hide selected source columns')
ColumnsSheet.addCommand(None, 'resize-source-rows-max', 'for c in selectedRows or [cursorRow]: c.setWidth(c.getMaxWidth(c.sheet.visibleRows))', 'adjust widths of selected source columns')

ColumnsSheet.addCommand('g%', 'type-float-selected', 'onlySelectedRows.type=float', 'set type of selected source columns to float')
ColumnsSheet.addCommand('g#', 'type-int-selected', 'onlySelectedRows.type=int', 'set type of selected source columns to int')
ColumnsSheet.addCommand('gz#', 'type-len-selected', 'onlySelectedRows.type=vlen', 'set type of selected source columns to len')
ColumnsSheet.addCommand('g@', 'type-date-selected', 'onlySelectedRows.type=date', 'set type of selected source columns to date')
ColumnsSheet.addCommand('g$', 'type-currency-selected', 'onlySelectedRows.type=currency', 'set type of selected source columns to currency')
ColumnsSheet.addCommand('g~', 'type-string-selected', 'onlySelectedRows.type=str', 'set type of selected source columns to str')
ColumnsSheet.addCommand('gz~', 'type-any-selected', 'onlySelectedRows.type=anytype', 'set type of selected source columns to anytype')
ColumnsSheet.addCommand('gz%', 'type-floatsi-selected', 'onlySelectedRows.type=floatsi', 'set type of selected source columns to floatsi')
ColumnsSheet.addCommand('', 'type-floatlocale-selected', 'onlySelectedRows.type=floatlocale', 'set type of selected source columns to float using system locale')

MetaSheet.options.header = 0


vd.addGlobals({
    'ColumnsSheet': ColumnsSheet,
    'MetaSheet': MetaSheet,
    'VisiDataMetaSheet': VisiDataMetaSheet,
})

vd.addMenuItems('''
    View > Columns > this sheet > columns-sheet
    View > Columns > all sheets > columns-all
''')



import collections
import math
from functools import partial
from visidata import DrawablePane, BaseSheet, vd, VisiData, CompleteKey, clipdraw, HelpSheet, colors, AcceptInput, AttrDict, drawcache_property, dispwidth


vd.theme_option('color_cmdpalette', 'black on 72', 'base color of command palette')
vd.theme_option('disp_cmdpal_max', 10, 'max number of suggestions for command palette')

vd.help_longname = '''# Choose Command
Start typing a command longname or keyword in its helpstring.

- `Enter` to execute top command.
- `Tab` to highlight top command.

## When Command Highlighted

- `Tab`/`Shift+Tab` to cycle highlighted command.
- `Enter` to execute highlighted command.
- `0-9` to execute numbered command.
'''

def add_to_input(v, i, value=''):
    items = list(v.split())
    if not v or v.endswith(' '):
        items.append(value)
    else:
        items[-1] = value
    v = ' '.join(items) + ' '
    return v, len(v)


def accept_input(v, i, value=None):
    raise AcceptInput(v if value is None else value)

def accept_input_if_subset(v, i, value=''):
    # if no input, accept value under cmd palette cursor
    if not v:
        raise AcceptInput(value)

    # if the last item is a partial match, replace it with the full value
    parts = v.split()
    if value and value.startswith(parts[-1]):
        v = ' '.join(parts[:-1] + [value])

    raise AcceptInput(v)

@VisiData.lazy_property
def usedInputs(vd):
    return collections.defaultdict(int)

@DrawablePane.after
def execCommand2(sheet, cmd, *args, **kwargs):
    vd.usedInputs[cmd.longname] += 1

@BaseSheet.api
def inputPalette(sheet, prompt, items,
                 value_key='key',
                 formatter=lambda m, item, trigger_key: f'{trigger_key} {item}',
                 multiple=False,
                 **kwargs):
    if not vd.wantsHelp('cmdpalette'):
        return vd.input(prompt,
                completer=CompleteKey(sorted(item[value_key] for item in items)),
                **kwargs)

    bindings = dict()

    #state variables for navigating display of matches
    prev_value = None
    tabitem = -1
    offset = 0
    def reset_display():
        nonlocal tabitem, offset
        tabitem = -1
        offset = 0

    def tab(n, nitems):
        nonlocal tabitem
        if not nitems: return None
        tabitem = (tabitem + n) % nitems

    def _draw_palette(value):
        nonlocal prev_value
        words = value.split()
        if value != prev_value:
            reset_display()
            prev_value = value

        if multiple and words:
            if value.endswith(' '):
                finished_words = words
                unfinished_words = []
            else:
                finished_words = words[:-1]
                unfinished_words = [words[-1]]
        else:
            finished_words = []
            unfinished_words = words

        unuseditems = [item for item in items if item[value_key] not in finished_words]

        matches = vd.fuzzymatch(unuseditems, unfinished_words)

        h = sheet.windowHeight
        w = min(100, sheet.windowWidth)
        nitems = min(h-2, sheet.options.disp_cmdpal_max)
        if nitems <= 0:
            return None

        useditems = []
        palrows = []
        n_results = 0
        def read_matches(offset):
            nonlocal useditems, palrows, value, n_results

            useditems = []
            palrows = []
            for m in matches[offset:offset+nitems]:
                useditems.append(m.match)
                palrows.append((m, m.match))
            n_results += len(matches)

            #List matches only, usually. But list the available choices when there's no input,
            #or (if multiple is True) they've just pressed space after a word.
            if not unfinished_words:
                favitems = sorted([item for item in unuseditems if item not in useditems],
                                key=lambda item: -vd.usedInputs.get(item[value_key], 0))
                for item in favitems[offset-len(palrows):offset+nitems-len(palrows)]:
                    palrows.append((None, item))
                n_results += len(favitems)
        read_matches(offset)

        def change_page(dir=+1):
            nonlocal offset, n_results, nitems
            new_offset = offset + dir*nitems
            # constrain offset to be a multiple of nitems
            new_offset = min(new_offset, ((n_results-1) // nitems)*nitems)
            new_offset = max(new_offset, 0)
            if new_offset == offset: return None
            offset = new_offset

        navailitems = min(len(palrows), nitems)

        bindings['^I'] = lambda *args: tab(1, navailitems) or args
        bindings['KEY_BTAB'] = lambda *args: tab(-1, navailitems) or args
        bindings['KEY_PPAGE'] = lambda *args: (change_page(-1) and read_matches(offset)) or args
        bindings['KEY_NPAGE'] = lambda *args: (change_page(+1) and read_matches(offset)) or args
        for numkey in '1234567890':
            bindings.pop(numkey, None)

        for i in range(nitems-len(palrows)):
            palrows.append((None, None))

        used_triggers = set()
        for i, (m, item) in enumerate(palrows):
            trigger_key = ''
            if tabitem >= 0 and item:
                tkey = f'{i+1}'[-1]
                if tkey not in used_triggers:
                    trigger_key = tkey
                    bindings[trigger_key] = partial(add_to_input if multiple else accept_input, value=item[value_key])
                    used_triggers.add(trigger_key)

            attr = colors.color_cmdpalette

            if tabitem < 0 and palrows:
                _ , topitem = palrows[0]
                if topitem:
                    if multiple:
                        bindings['^J'] = partial(accept_input_if_subset, value=topitem[value_key])
                        bindings[' '] = partial(add_to_input, value=topitem[value_key])
                    else:
                        bindings['^J'] = partial(accept_input, value=topitem[value_key])
            elif item and i == tabitem:
                if multiple:
                    bindings['^J'] = partial(accept_input_if_subset, value=item[value_key])
                    bindings[' '] = partial(add_to_input, value=item[value_key])
                else:
                    bindings['^J'] = partial(accept_input, value=item[value_key])
                attr = colors.color_menu_spec

            match_summary = formatter(m, item, trigger_key) if item else ' '

            clipdraw(sheet._scr, h-nitems-2+i, 0, match_summary, attr, w=w)
        attr = colors.color_cmdpalette
        instr = 'Press [:keystrokes]PgUp/PgDn[/] to scroll items, [:keystrokes]Tab/Shift+Tab/Enter[/] to choose, [:keystrokes]Esc[/] to cancel.'
        clipdraw(sheet._scr, h-2, 0, instr, attr, w=w)

        return None

    completer = CompleteKey(sorted(item[value_key] for item in items))
    return vd.input(prompt,
            completer=completer,
            updater=_draw_palette,
            bindings=bindings,
            **kwargs)


def cmdlist(sheet):
    return [
            AttrDict(longname=row.longname,
                     description=sheet.cmddict[(row.sheet, row.longname)].helpstr)
        for row in sheet.rows
    ]
HelpSheet.cmdlist = drawcache_property(cmdlist)


@BaseSheet.api
def inputLongname(sheet):
    prompt = 'command name: '
    # get set of commands possible in the sheet
    this_sheets_help = HelpSheet('', source=sheet)
    vd.sync(this_sheets_help.ensureLoaded())

    def _fmt_cmdpal_summary(match, row, trigger_key):
        keystrokes = this_sheets_help.revbinds.get(row.longname, [None])[0] or ' '
        formatted_longname = match.formatted.get('longname', row.longname) if match else row.longname
        formatted_name = f'[:bold][:onclick {row.longname}]{formatted_longname}[/][/]'
        if vd.options.debug and match:
            keystrokes = f'[{match.score}]'
        r = f' [:keystrokes]{keystrokes.rjust(dispwidth(prompt)-5)}[/]  '
        if trigger_key:
            r += f'[:keystrokes]{trigger_key}[/]'
        else:
            r += ' '

        r += f' {formatted_name}'
        if row.description:
            formatted_desc = match.formatted.get('description', row.description) if match else row.description
            r += f' - {formatted_desc}'
        return r

    return sheet.inputPalette(prompt, this_sheets_help.cmdlist,
                              value_key='longname',
                              formatter=_fmt_cmdpal_summary,
                              help=vd.help_longname,
                              type='longname')

@BaseSheet.api
def inputLongnameSimple(sheet):
    'Input a command longname without using the command palette.'
    longnames = set(k for (k, obj), v in vd.commands.iter(sheet))
    return vd.input("command name: ", completer=CompleteKey(sorted(longnames)), type='longname')


@BaseSheet.api
def exec_longname(sheet, longname):
    if not sheet.getCommand(longname):
        vd.fail(f'no command {longname}')
    sheet.execCommand(longname)


vd.addCommand('Space', 'exec-longname', 'exec_longname(inputLongname())', 'execute command by its longname')
vd.addCommand('zSpace', 'exec-longname-simple', 'exec_longname(inputLongnameSimple())', 'execute command by its longname (without command palette)')



from copy import copy
import collections
import string
import itertools
import threading
import re
import time
import json

from visidata import options, anytype, stacktrace, vd, drawcache
from visidata import asyncthread, dispwidth, clipstr, iterchars
from visidata import wrapply, TypedWrapper, TypedExceptionWrapper
from visidata import Extensible, AttrDict, undoAttrFunc, ExplodingMock, MissingAttrFormatter
from visidata import getitem, setitem, getitemdef, getitemdeep, setitemdeep, getattrdeep, setattrdeep, iterchunks

class InProgress(Exception):
    @property
    def stacktrace(self):
        return  ['calculation in progress']

INPROGRESS = TypedExceptionWrapper(None, exception=InProgress())  # sentinel

vd.option('col_cache_size', 0, 'max number of cache entries in each cached column')
vd.option('disp_formatter', 'generic', 'formatter to create the text in each cell (also used by text savers)', replay=True)
vd.option('disp_displayer', 'generic', 'displayer to render the text in each cell', replay=False)


class DisplayWrapper:
    def __init__(self, value=None, *, typedval=None, text='', note=None, notecolor=None, error=None):
        self.value = value      # actual value (any type)
        self.typedval = typedval  # consistently typed value (or None)
        self.text = text  # displayed string
        self.note = note        # single unicode character displayed in cell far right
        self.notecolor = notecolor  # configurable color name (like 'color_warning')
        self.error = error      # list of strings for stacktrace

    def __bool__(self):
        return bool(self.value)

    def __eq__(self, other):
        return self.value == other


class Column(Extensible):
    '''Base class for all column types.

        - *name*: name of this column; if None, current sheet will assign a name
        - *type*: ``anytype str int float date`` or other type-like conversion function.
        - *cache*: cache behavior

           - ``False`` (default): getValue never caches; calcValue is always called.
           - ``True``: getValue maintains a cache of ``options.col_cache_size``.
           - ``"async"``: ``getValue`` launches thread for every uncached result, returns invalid value until cache entry available.
        - *width*: == 0 if hidden, None if auto-compute next time.
        - *height*: max height, None/0 to auto-compute for each row.
        - *fmtstr*: format string as applied by column type.
        - *getter*: default calcValue calls ``getter(col, row)``.
        - *setter*: default putValue calls ``setter(col, row, val)``.
        - *kwargs*: other attributes to be set on this column.
    '''
    def __init__(self, name=None, *, type=anytype, cache=False, **kwargs):
        self.sheet = ExplodingMock('use addColumn() on all columns')  # owning Sheet, set in .recalc() via Sheet.addColumn
        if name is None:
            if vd.sheet: # get a column name from the current sheet
                name = vd.sheet.incremented_colname()
            else:
                name = ''
        self.name = str(name) # display visible name
        self.fmtstr = ''      # by default, use str()
        self._type = type     # anytype/str/int/float/date/func
        self.getter = lambda col, row: row
        self.setter = None
        self._width = None    # == 0 if hidden, None if auto-compute next time
        self.hoffset = 0      # starting horizontal (char) offset of displayed column value
        self.voffset = 0      # starting vertical (line) offset of displayed column value
        self.height = 1       # max height, None/0 to auto-compute for each row
        self.keycol = 0       # keycol index (or 0 if not key column)
        self.expr = None      # Column-type-dependent parameter
        self.formatter = ''
        self.displayer = ''
        self.defer = False
        self.disp_expert = 0    # do not show if 'nometacols' in options.disp_help_flags

        self.setCache(cache)
        for k, v in kwargs.items():
            setattr(self, k, v)  # instead of __dict__.update(kwargs) to invoke property.setters

    def __copy__(self):
        cls = self.__class__
        ret = cls.__new__(cls)
        ret.__dict__.update(self.__dict__)
        ret.keycol = 0   # column copies lose their key status
        if self._cachedValues is not None:
            ret._cachedValues = collections.OrderedDict()  # an unrelated cache for copied columns
        return ret

    def __str__(self):
        return f'{type(self).__name__}:{self.name}'

    def __repr__(self):
        return f'<{type(self).__name__}: {self.name}>'

    def __deepcopy__(self, memo):
        return self.__copy__()  # no separate deepcopy

    def __getstate__(self):
        return {k:getattr(self, k) for k in 'name typestr width height expr keycol formatter displayer fmtstr voffset hoffset aggstr'.split() if hasattr(self, k)}

    def __setstate__(self, d):
        for attr, v in d.items():
            setattr(self, attr, v)

    def recalc(self, sheet=None):
        'Reset column cache, attach column to *sheet*, and reify column name.'
        if self._cachedValues:
            self._cachedValues.clear()
        if sheet:
            self.sheet = sheet
        self.name = self._name

    @property
    def name(self):
        'Name of this column.'
        return self._name or ''

    @name.setter
    def name(self, name):
        self.setName(name)

    def setName(self, name):
        if name is None:
            name = ''
        if isinstance(name, str):
            name = name.strip()
        else:
            name = str(name)

        if self.sheet:
            name = self.sheet.maybeClean(name)

        self._name = name

    @property
    def typestr(self):
        'Type of this column as string.'
        return self._type.__name__

    @typestr.setter
    def typestr(self, v):
        self.type = vd.getGlobals()[v or 'anytype']

    @property
    def type(self):
        'Type of this column.'
        return self._type

    @type.setter
    def type(self, t):
        if self._type != t:
            vd.addUndo(setattr, self, '_type', self.type)
        if not t:
            self._type = anytype
        elif isinstance(t, str):
            self.typestr = t
        else:
            self._type = t

    @property
    def width(self):
        'Width of this column in characters.  0 or negative means hidden.  None means not-yet-autocomputed.'
        return self._width

    @width.setter
    def width(self, w):
        if self.width != w:
            if self.width == 0 or w == 0:  # hide/unhide
                vd.addUndo(setattr, self, '_width', self.width)
            self._width = w

    @property
    def formatted_help(self):
        return MissingAttrFormatter().format(self.help, sheet=self.sheet, col=self, vd=vd)

    @property
    def help_formatters(self):
        formatters = [k[10:] for k in dir(self) if k.startswith('formatter_')]
        return ' '.join(formatters)

    @property
    def help_displayers(self):
        displayers = [k[10:] for k in dir(self) if k.startswith('displayer_')]
        return ' '.join(displayers)

    @property
    def _formatdict(col):
        if '=' in col.fmtstr:
            return dict(val.split('=', maxsplit=1) for val in col.fmtstr.split())
        return {}

    @property
    def fmtstr(self):
        'Format string to use to display this column.'
        return self._fmtstr or vd.getType(self.type).fmtstr

    @fmtstr.setter
    def fmtstr(self, v):
        self._fmtstr = v

    def _format_len(self, typedval, **kwargs):
        if isinstance(typedval, dict):
            return f'{{{len(typedval)}}}'
        elif isinstance(typedval, (list, tuple)):
            return f'[{len(typedval)}]'

        return self.formatValue(typedval, **kwargs)

    def formatter_len(self, fmtstr):
        return self._format_len

    def formatter_generic(self, fmtstr):
        return self.formatValue

    def formatter_json(self, fmtstr):
        return lambda v,*args,**kwargs: json.dumps(v)

    def formatter_python(self, fmtstr):
        return lambda v,*args,**kwargs: str(v)

    @drawcache
    def make_formatter(self):
        'Return function for format(v) from the current formatter and fmtstr'
        _formatMaker = getattr(self, 'formatter_'+(self.formatter or self.sheet.options.disp_formatter))
        return _formatMaker(self._formatdict)

    def format(self, *args, **kwargs):
        return self.make_formatter()(*args, **kwargs)

    def formatValue(self, typedval, width=None):
        '''Return displayable string of *typedval* according to ``Column.fmtstr``.
        If *width* is not None, values are clipped to that width when *typedval*
        is a dict/list/tuple, but not for other types.'''
        if typedval is None:
            return None

        if self.type is anytype:
            if isinstance(typedval, (dict, list, tuple)):
                if width is None:
                    return ''.join(iterchars(typedval))
                dispval, dispw = clipstr(iterchars(typedval), width)
                return dispval

        if isinstance(typedval, bytes):
            typedval = typedval.decode(options.encoding, options.encoding_errors)

        gt = vd.getType(self._type)
        return gt.formatter(self._fmtstr or gt.fmtstr, typedval)

    def displayer_generic(self, dw:DisplayWrapper, width=None):
        '''Fit *dw.text* into *width* charcells.
           Generate list of (attr:str, text:str) suitable for clipdraw_chunks.

           The 'generic' displayer does not do any formatting.
        '''
        if width is not None and width > 1 and \
                vd.isNumeric(self) and \
                isinstance(dw.typedval, (int, float)):
            yield ('', dw.text.rjust(width-2))
        else:
            yield ('', dw.text)

    def displayer_full(self, dw:DisplayWrapper, width=None):
        '''Fit *dw.text* into *width* charcells.
           Generate list of (attr:str, text:str) suitable for clipdraw_chunks.

           The 'full' displayer allows formatting like [:color].
        '''
        if width is not None and width > 1 and vd.isNumeric(self):
            yield from iterchunks(dw.text.rjust(width-2))
        else:
            yield from iterchunks(dw.text)

    def display(self, *args, **kwargs):
        f = getattr(self, 'displayer_'+(self.displayer or self.sheet.options.disp_displayer), self.displayer_generic)
        return f(*args, **kwargs)

    def hide(self, hide=True):
        if hide:
            self.setWidth(0)
        else:
            self.setWidth(abs(self.width or self.getMaxWidth(self.sheet.visibleRows)))

    @property
    def hidden(self):
        'Return True if width of this column is 0 or negative.'
        if self.width is None:
            return False
        return self.width <= 0

    def calcValue(self, row):
        'Calculate and return value for *row* in this column.'
        return (self.getter)(self, row)

    def getTypedValue(self, row):
        'Return the properly-typed value for the given row at this column, or a TypedWrapper object in case of null or error.'
        return wrapply(self.type, wrapply(self.getValue, row))

    def setCache(self, cache):
        '''Set cache behavior for this column to *cache*:

           - ``False`` (default): getValue never caches; calcValue is always called.
           - ``True``: getValue maintains a cache of ``options.col_cache_size``.
           - ``"async"``: ``getValue`` launches thread for every uncached result, maintains cache of infinite size.  Returns invalid value until cache entry available.'''
        self.cache = cache
        self._cachedValues = collections.OrderedDict() if self.cache else None

    @asyncthread
    def _calcIntoCacheAsync(self, row):
        # causes issues when moved into _calcIntoCache gen case
        self._cachedValues[self.sheet.rowid(row)] = INPROGRESS
        self._calcIntoCache(row)

    def _calcIntoCache(self, row):
        ret = wrapply(self.calcValue, row)
        if not isinstance(ret, TypedExceptionWrapper) or ret.val is not INPROGRESS:
            self._cachedValues[self.sheet.rowid(row)] = ret
        return ret

    def getValue(self, row):
        'Return value for *row* in this column, calculating if not cached.'

        if self.defer:
            try:
                row, rowmods = self.sheet._deferredMods[self.sheet.rowid(row)]
                return rowmods[self]
            except KeyError:
                pass

        if self._cachedValues is None:
            return self.calcValue(row)

        k = self.sheet.rowid(row)
        if k in self._cachedValues:
            return self._cachedValues[k]

        if self.cache == 'async':
            ret = self._calcIntoCacheAsync(row)
        else:
            ret = self._calcIntoCache(row)

            cachesize = options.col_cache_size
            if cachesize > 0 and len(self._cachedValues) > cachesize:
                self._cachedValues.popitem(last=False)

        return ret

    def getCell(self, row):
        '''Return DisplayWrapper for displayable cell value.
        For dict/list/tuple cells, the width of the value returned is capped at the column width.'''
        cellval = wrapply(self.getValue, row)
        typedval = wrapply(self.type, cellval)

        if isinstance(typedval, TypedWrapper):
            if isinstance(cellval, TypedExceptionWrapper):  # calc failed
                exc = cellval.exception
                if cellval.forwarded:
                    dispval = str(cellval)  # traceback.format_exception_only(type(exc), exc)[-1].strip()
                else:
                    dispval = options.disp_error_val
                return DisplayWrapper(cellval.val, error=exc.stacktrace,
                                        text=dispval,
                                        note=f'[:onclick error-cell]{options.disp_note_getexc}[:]',
                                        notecolor='color_error')
            elif typedval.val is None:  # early out for strict None
                return DisplayWrapper(None, text='',  # force empty display for None
                                            note=options.disp_note_none,
                                            notecolor='color_note_type')
            elif isinstance(typedval, TypedExceptionWrapper):  # calc succeeded, type failed
                return DisplayWrapper(typedval.val, text=str(cellval),
                                            error=typedval.stacktrace,
                                            note=f'[:onclick error-cell]{options.disp_note_typeexc}[:]',
                                            notecolor='color_warning')
            else:
                return DisplayWrapper(typedval.val, text=str(typedval.val),
                                            error=['unknown'],
                                            note=f'[:onclick error-cell]{options.disp_note_typeexc}[:]',
                                            notecolor='color_warning')
        elif isinstance(typedval, threading.Thread):
            return DisplayWrapper(None,
                                text=options.disp_pending,
                                note=options.disp_note_pending,
                                notecolor='color_note_pending')

        dw = DisplayWrapper(cellval)
        dw.typedval = typedval

        try:
            dw.text = self.format(typedval, width=self.width) or ''

            # annotate cells with raw value type in anytype columns, except for strings
            if self.type is anytype and type(cellval) is not str:
                typedesc = vd.typemap.get(type(cellval), None)
                if typedesc:
                    dw.note = typedesc.icon
                    dw.notecolor = 'color_note_type'

        except Exception as e:  # formatting failure
            e.stacktrace = stacktrace()
            dw.error = e.stacktrace
            try:
                dw.text = str(cellval)
            except Exception as e:
                dw.text = str(e)
            dw.note = options.disp_note_fmtexc
            dw.notecolor = 'color_warning'

#        dw.display = self.display(dw)   # set during draw() once colwidth is known
        return dw

    def getDisplayValue(self, row):
        '''Return string displayed in this column for given *row*.
        For dict/list/tuple cells, the width of the display value returned is capped at the column width.'''
        return self.getCell(row).text

    def putValue(self, row, val):
        'Change value for *row* in this column to *val* immediately.  Does not check the type.  Overridable; by default calls ``.setter(row, val)``.'
        if self.setter:
            return self.setter(self, row, val)

    def setValue(self, row, val, setModified=True):
        'Change value for *row* in this column to *val*.  Call ``putValue`` immediately if not a deferred column (added to deferred parent at load-time); otherwise cache until later ``putChanges``.  Caller must add undo function.'
        if self.defer:
            self.cellChanged(row, val)
        else:
            self.putValue(row, val)
        if setModified:  #1800
            self.sheet.setModified()

    @asyncthread
    def setValues(self, rows, *values):
        'Set values in this column for *rows* to *values*, recycling values as needed to fill *rows*.'
        vd.addUndoSetValues([self], rows)

        for r, v in zip(rows, itertools.cycle(values)):
            vd.callNoExceptions(self.setValue, r, v)

        self.recalc()
        return vd.status('set %d cells to %d values' % (len(rows), len(values)))

    @asyncthread
    def setValuesTyped(self, rows, *values):
        'Set values on this column for *rows* to *values*, coerced to column type, recycling values as needed to fill *rows*.  Abort on type exception.'
        vd.addUndoSetValues([self], rows)
        for r, v in zip(rows, itertools.cycle(self.type(val) for val in values)):
            vd.callNoExceptions(self.setValue, r, v)

        self.recalc()

        return vd.status('set %d cells to %d values' % (len(rows), len(values)))

    def getMaxWidth(self, rows):
        'Return the maximum length of any cell in column or its header (up to drawable window width).'
        drawable_width = self.sheet.windowWidth-1
        nlen = dispwidth(self.name)
        w_max = nlen
        for r in rows:
            row_w = self.measureValueWidthCapped(r, maxwidth=drawable_width)
            if w_max < row_w:
                w_max = row_w
            if w_max >= self.sheet.windowWidth:
                break  #1747  early out to speed up wide columns
        return min(w_max+2, drawable_width)

    def measureValueWidthCapped(self, row, maxwidth=None):
        '''Measure the width of the contents of a cell. Stop measuring at *maxwidth*,
           to save time iterating over very long dict/list/tuple values.
           If *maxwidth* is None, return the full width.'''
        # The value classification logic here is taken from getCell,
        # modified to cap the width examined for any dict/list/tuple
        cellval = wrapply(self.getValue, row)
        typedval = wrapply(self.type, cellval)
        if isinstance(typedval, (TypedWrapper, threading.Thread)):
            return dispwidth(self.getCell(row).text, maxwidth=maxwidth)
        try:
            text = self.format(typedval, width=maxwidth) or ''
        except Exception as e:  # formatting failure
            try:
                text = str(cellval)
            except Exception as e:
                text = str(e)
        return dispwidth(text, maxwidth=maxwidth)


# ---- basic Columns

class AttrColumn(Column):
    'Column using getattr/setattr with *attr*.'
    def __init__(self, name=None, expr=None, **kwargs):
        super().__init__(name,
            expr=expr if expr is not None else name,
            getter=lambda col,row: getattrdeep(row, col.expr, None),
            **kwargs)

    def putValue(self, row, val):
        super().putValue(row, val)
        setattrdeep(row, self.expr, val)


class ItemColumn(Column):
    'Column using getitem/setitem with *expr*.'
    def __init__(self, name=None, expr=None, **kwargs):
        super().__init__(name,
            expr=expr if expr is not None else name,
            getter=lambda col,row: getitemdeep(row, col.expr, None),
            **kwargs)

    def putValue(self, row, val):
        super().putValue(row, val)
        setitemdeep(row, self.expr, val)


class SubColumnFunc(Column):
    'Column compositor; preprocess row with *subfunc*(row, *expr*) before passing to *origcol*.getValue and *origcol*.setValue.'
    def __init__(self, name='', origcol=None, expr=None, subfunc=getitemdef, **kwargs):
        super().__init__(name, type=origcol.type, width=origcol.width, expr=expr, **kwargs)
        self.origcol = origcol
        self.subfunc = subfunc

    def calcValue(self, row):
        subrow = self.subfunc(row, self.expr)
        if subrow is not None:
            # call getValue to use deferred values from source sheet
            return self.origcol.getValue(subrow)

    def putValue(self, row, value):
        subrow = self.subfunc(row, self.expr)
        if subrow is None:
            vd.fail('no source row')
        self.origcol.setValue(subrow, value)

    def recalc(self, sheet=None):
        Column.recalc(self, sheet)
        self.origcol.recalc()  # reset cache but don't change sheet


def SubColumnAttr(attrname, c, **kwargs):
    if 'name' not in kwargs:
        kwargs['name'] = c.name
    return SubColumnFunc(origcol=c, subfunc=getattrdeep, expr=attrname, **kwargs)

def SubColumnItem(idx, c, **kwargs):
    if 'name' not in kwargs:
        kwargs['name'] = c.name
    return SubColumnFunc(origcol=c, subfunc=getitemdef, expr=idx, **kwargs)


class SettableColumn(Column):
    'Column using rowid to store and retrieve values internally.'
    def putValue(self, row, value):
        self._store[self.sheet.rowid(row)] = value

    def calcValue(self, row):
        return self._store.get(self.sheet.rowid(row), None)


SettableColumn.init('_store', dict, copy=True)


vd.addGlobals(
    INPROGRESS=INPROGRESS,
    Column=Column,
    setitem=setitem,
    getattrdeep=getattrdeep,
    setattrdeep=setattrdeep,
    getitemdef=getitemdef,
    AttrColumn=AttrColumn,
    ItemColumn=ItemColumn,
    SettableColumn=SettableColumn,
    SubColumnFunc=SubColumnFunc,
    SubColumnItem=SubColumnItem,
    SubColumnAttr=SubColumnAttr,

# synonyms
    ColumnItem=ItemColumn,
    ColumnAttr=AttrColumn,
    DisplayWrapper=DisplayWrapper
)



'''Loader for the jrnl.sh CLI journal file format'''

import re


from visidata import VisiData, TableSheet, ItemColumn, AttrDict


@VisiData.api
def open_jrnl(vd, p):
    return JrnlSheet(p.base_stem, source=p)


class JrnlSheet(TableSheet):
    # rowdef: AttrDict
    columns = [
        ItemColumn('date'),
        ItemColumn('time'),
        ItemColumn('title'),
        ItemColumn('body'),
        ItemColumn('tags'),
    ]
    def iterload(self):
        re_title = re.compile(r'\[(.*?)\s(.*?)\] (.*)')
        prevline = ''
        for line in self.source:
            tags = re.findall(r'(?<!\S)(@[-+*#/\w]+)', line)
            if not prevline:
                m = re_title.match(line)
                if m:
                    row = AttrDict()
                    row.date, row.time, row.title = m.groups()
                    row.body = ''
                    row.tags = ' '.join(tags)
                    yield row
                    continue

            row.body += line + '\n'
            row.tags = ' '.join([row.tags]+tags)
            prevline = line.strip()


@VisiData.api
def save_jrnl(vd, p, *vsheets):
    with p.open(mode='w', encoding=vsheets[0].options.save_encoding) as fp:
        for vs in vsheets:
            for r in vs.iterrows():
                fp.write(f'[{r.date} {r.time}] {r.title}\n')
                body = r.body.strip()
                if body:
                    fp.write(body + '\n')
                fp.write('\n')



'''
Status messages get added with vd.{debug/aside/status/warning/fail/error}(), and cleared in mainloop
'''

import builtins
import collections
import curses
import sys

import visidata
from visidata import vd, VisiData, BaseSheet, Sheet, ColumnItem, Column, RowColorizer, options, colors, wrmap, clipdraw, ExpectedException, update_attr, dispwidth, ColorAttr, clipstr_middle



vd.option('disp_rstatus_fmt', '{sheet.threadStatus} {sheet.keystrokeStatus}   [:longname_status]{sheet.longname}[/]  {sheet.nRows:9d} {sheet.rowtype} {sheet.modifiedStatus}{sheet.selectedStatus}{vd.replayStatus}{vd.sidebarStatus}', 'right-side status format string')
vd.option('disp_status_fmt', '{sheet.sheetlist}| ', 'left-side status format string')
vd.theme_option('disp_lstatus_max', 0, 'maximum length of left status line')
vd.theme_option('disp_status_sep', '│', 'separator between statuses')

vd.theme_option('color_keystrokes', 'bold white on 237', 'color of input keystrokes')
vd.theme_option('color_longname_guide', '237', 'color of command longnames')
vd.theme_option('color_longname_status', 'white', 'color of command longnames')
vd.theme_option('color_keys', 'bold reverse', 'color of keystrokes in help')
vd.theme_option('color_status', 'bold on 238', 'status line color')
vd.theme_option('color_error', '202 1', 'error message color')
vd.theme_option('color_warning', '166 15', 'warning message color')
vd.theme_option('color_top_status', 'underline', 'top window status bar color')
vd.theme_option('color_active_status', 'black on 68 blue', ' active window status bar color')
vd.theme_option('color_inactive_status', '8 on black', 'inactive window status bar color')
vd.theme_option('color_highlight_status', 'black on green', 'color of highlighted elements in statusbar')

BaseSheet.init('longname', lambda: '')

@BaseSheet.property
def ancestors(sheet):
    if isinstance(sheet.source, BaseSheet):
        return sheet.source.ancestors + [sheet.source]
    else:
        return []

@BaseSheet.property
def sheetlist(sheet):
    leafsheets = []
    parents = set()

    sheetstack = vd.sheetstack(sheet.pane)
    sheets = [x for x in vd.allSheets if x in sheetstack]+ [x for x in sheetstack if x not in vd.allSheets]

    sheetnames = []
    for vs in sheets:
        if not vs.precious:  #2573
            continue
        if isinstance(vs, BaseSheet):
            shortcut = ' '
            if vs.shortcut in '1 2 3 4 5 6 7 8 9 10'.split():
                shortcut = vs.shortcut[-1] + vs.icon
            if vs is vd.sheet:
                sheetnames.append(f'[:menu_active]{shortcut}{vs.name}[:]')
            else:
                sheetnames.append(f'[:onclick jump-sheet-{vs.shortcut}]' + clipstr_middle(f'{shortcut}{vs.name}', 20)[0] + '[:]')
        else:
            sheetnames.append(vs)

    return ' | '.join(sheetnames)

@BaseSheet.api
def _updateStatusBeforeExec(sheet, cmd, args, ks):
    sheet.longname = cmd.longname
    if sheet._scr:
        vd.drawRightStatus(sheet._scr, sheet)  #996 show longname during commands
        sheet._scr.refresh()


vd.beforeExecHooks.append(BaseSheet._updateStatusBeforeExec)


@VisiData.lazy_property
def statuses(vd):
    return collections.OrderedDict()  # (priority, statusmsg) -> num_repeats; shown until next action


@VisiData.lazy_property
def statusHistory(vd):
    return list()  # list of [priority, statusmsg, repeats] for all status messages ever

@VisiData.api
def getStatusSource(vd):
    return None


@VisiData.api
def status(vd, *args, priority=0):
    'Display *args* on status until next action.'
    if not args:
        return True

    k = (priority, tuple(map(str, args)))
    vd.statuses[k] = vd.statuses.get(k, 0) + 1

    source = vd.getStatusSource()

    if not vd.cursesEnabled:
        msg = '\r' + composeStatus(args)
        if vd.options.debug:
            msg += f' [{source}]'
        builtins.print(msg, file=sys.stderr)

    return vd.addToStatusHistory(*args, priority=priority, source=source)

@VisiData.api
def addToStatusHistory(vd, *args, priority=0, source=None):
    if vd.statusHistory:
        prevpri, prevargs, _, _ = vd.statusHistory[-1]
        if prevpri == priority and prevargs == args:
            vd.statusHistory[-1][2] += 1
            return True

    vd.statusHistory.append([priority, args, 1, source])
    return True

@VisiData.api
def error(vd, *args):
    'Abort with ExpectedException, and display *args* on status as an error.'
    vd.status(*args, priority=3)
    raise ExpectedException(args[0] if args else '')

@VisiData.api
def fail(vd, *args):
    'Abort with ExpectedException, and display *args* on status as a warning.'
    vd.status(*args, priority=2)
    raise ExpectedException(args[0] if args else '')

@VisiData.api
def warning(vd, *args):
    'Display *args* on status as a warning.'
    vd.status(*args, priority=1)

@VisiData.api
def aside(vd, *args, priority=0):
    'Add a message to statuses without showing the message proactively.'
    return vd.addToStatusHistory(*args, priority=priority, source=vd.getStatusSource())

@VisiData.api
def debug(vd, *args, **kwargs):
    'Display *args* on status if options.debug is set.'
    if options.debug:
        return vd.status(*args, **kwargs)

def middleTruncate(s, w):
    if len(s) <= w:
        return s
    return s[:w] + options.disp_truncator + s[-w:]


def composeStatus(msgparts, n=1):
    msg = '; '.join(wrmap(str, msgparts))
    if n > 1:
        msg = '[%sx] %s' % (n, msg)
    return msg


@BaseSheet.api
def leftStatus(sheet):
    'Return left side of status bar for this sheet. Overridable.'
    return sheet.formatString(sheet.options.disp_status_fmt)


@VisiData.api
def drawLeftStatus(vd, scr, vs):
    'Draw left side of status bar.'
    cattr = colors.get_color('color_active_status')
    active = (vs is vd.activeSheet)
    if active:
        cattr = update_attr(cattr, colors.color_active_status, 1)
    else:
        cattr = update_attr(cattr, colors.color_inactive_status, 1)

    if scr is vd.winTop:
        cattr = update_attr(cattr, colors.color_top_status, 1)

    x = 0
    y = vs.windowHeight-1  # status for each window
    lstatus = vs.leftStatus()
    maxwidth = options.disp_lstatus_max
    if maxwidth > 0:
        lstatus = middleTruncate(lstatus, maxwidth//2)

    x = clipdraw(scr, y, 0, lstatus, cattr, w=vs.windowWidth-1)

    vd.onMouse(scr, 0, y, x, 1,
                    BUTTON3_PRESSED='rename-sheet',
                    BUTTON3_CLICKED='rename-sheet')


@VisiData.api
def rightStatus(vd, sheet):
    'Return right side of status bar.  Overridable.'
    return sheet.formatString(sheet.options.disp_rstatus_fmt)


@BaseSheet.property
def keystrokeStatus(vs):
    if vs is vd.activeSheet:
        return f'[:keystrokes]{vd.keystrokes}[/]'

    return ''


@BaseSheet.property
def threadStatus(vs) -> str:
    if vs.currentThreads:
        ret = str(vd.checkMemoryUsage())
        gerunds = [p.gerund for p in vs.progresses if p.gerund] or ['processing']
        ret += f' [:working]{vs.progressPct} {gerunds[0]}…[/]'
        return ret
    return ''

@BaseSheet.property
def modifiedStatus(sheet):
    ret = ' [M]' if sheet.hasBeenModified else ''
    if not vd.couldOverwrite():
        ret += ' [:highlight_status][RO][/] '
    return ret


@BaseSheet.property
def selectedStatus(sheet):
    return ''


@Sheet.property
def selectedStatus(sheet):
    if sheet.nSelectedRows:
        return f' [:selected_row][:onclick dup-selected]{sheet.options.disp_selected_note}{sheet.nSelectedRows}[/][/] '

@VisiData.api
def drawRightStatus(vd, scr, vs):
    'Draw right side of status bar.  Return length displayed.'
    rightx = vs.windowWidth

    statuslen = 0
    try:
        cattr = ColorAttr()
        if scr is vd.winTop:
            cattr = update_attr(cattr, colors.color_top_status, 0)
        cattr = update_attr(cattr, colors.color_active_status if vs is vd.activeSheet else colors.color_inactive_status, 0)
        rstat = vd.rightStatus(vs)
        x = max(2, rightx-dispwidth(rstat)-1)
        statuslen = clipdraw(scr, vs.windowHeight-1, x, rstat, cattr, w=vs.windowWidth-1)
    finally:
        if scr:
            curses.doupdate()
    return statuslen


class StatusSheet(Sheet):
    precious = False
    rowtype = 'statuses'  # rowdef: (priority, args, nrepeats)
    columns = [
        ColumnItem('priority', 0, type=int, width=0),
        ColumnItem('nrepeats', 2, type=int, width=0),
        ColumnItem('args', 1, width=0),
        Column('message', width=50, getter=lambda col,row: composeStatus(row[1], row[2])),
        ColumnItem('source', 3, width=0),
    ]
    colorizers = [
        RowColorizer(1, 'color_error', lambda s,c,r,v: r and r[0] == 3),
        RowColorizer(1, 'color_warning', lambda s,c,r,v: r and r[0] in [1,2]),
    ]

    def reload(self):
        self.rows = self.source


@VisiData.property
def statusHistorySheet(vd):
    return StatusSheet("status_history", source=vd.statusHistory[::-1])  # in reverse order


BaseSheet.addCommand('^P', 'open-statuses', 'vd.push(vd.statusHistorySheet)', 'open Status History')

vd.addMenuItems('''
    View > Statuses > open-statuses
''')



import functools
from visidata import clipdraw, colors, BaseSheet, VisiData, VisiDataMetaSheet, vd, EscapeException, AttrDict, ENTER, dispwidth, AcceptInput


@VisiData.api
def open_mnu(vd, p):
    return FormSheet(p.base_stem, source=p)


vd.save_mnu=vd.save_tsv

class FormSheet(VisiDataMetaSheet):
    rowtype='labels' # rowdef: { .x .y .text .color .command .input .key}

@VisiData.api
def replayCommand(vd, longname, input=None, sheet=None, col='', row=''):
    vd.replayOne(vd.cmdlog.newRow(sheet=sheet, col=col, row=row, longname=longname, input=input))


class FormCanvas(BaseSheet):
    rowtype = 'labels'
    pressedLabel = None

    def onPressed(self, r):
        if r.key:
            self.pressedLabel = r

    def onReleased(self, r):
        if self.pressedLabel is r:
            raise AcceptInput(r.key)

    def reload(self):
        self.rows = self.source.rows

    def draw(self, scr):
        vd.clearCaches()
        h, w = scr.getmaxyx()
        for r in self.source.rows:
            if not r.text:
                continue
            x, y = r.x, r.y
            if isinstance(y, float) and (0 < y < 1) or (-1 < y < 0): y = h*y
            if isinstance(x, float) and (0 < x < 1) or (-1 < x < 0): x = w*x-(dispwidth(r.text)/2)
            x = int(x)
            y = int(y)
            if y < 0: y += h
            if x < 0: x += w
            color = r.color
            if r is self.pressedLabel:
                color += ' reverse'
            clipdraw(scr, y, x, r.text, colors[color])
            # underline first occurrence of r.key in r.text
            if hasattr(r, 'key') and r.key:
                index = r.text.find(r.key)
                clipdraw(scr, y, x+index, r.text[index:index+len(r.key)], colors[color + " underline"])
            vd.onMouse(scr, x, y, dispwidth(r.text), 1,
                    BUTTON1_PRESSED=lambda y,x,key,r=r,sheet=self: sheet.onPressed(r),
                    BUTTON1_RELEASED=lambda y,x,key,r=r,sheet=self: sheet.onReleased(r))

    def run(self, scr):
        vd.setWindows(vd.scrFull)
        drawnrows = [r for r in self.source.rows if r.text]
        inputs = [r for r in self.source.rows if r.input]
        maxw = max(int(r.x)+dispwidth(r.text) for r in drawnrows)
        maxh = max(int(r.y) for r in drawnrows)
        h, w = vd.scrFull.getmaxyx()
        y, x = max(0, (h-maxh)//2-1), max(0, (w-maxw)//2-1)
        self.scrForm = vd.subwindow(vd.scrFull, x, y, min(w-1, maxw+1), min(h-1, maxh+2))
        self.scrForm.keypad(1)
        curinput = inputs[0] if inputs else None
        vd.draw_all()

        self.scrForm.erase()
        self.scrForm.border()
        self.draw(self.scrForm)

        while True:
            k = vd.getkeystroke(self.scrForm, self)
            if k in ['^C', '^Q', '^[', 'q']:
                return {}
            if curinput and k in curinput.keystrokes:
                return {curinput.input: k}
            if k == 'KEY_MOUSE':
                r = vd.parseMouse(form=self.scrForm)
                if r.found:
                    f = vd.getMouse(r.x, r.y, r.keystroke)
                    if f:
                        try:
                            f(r.y, r.x, r.keystroke)
                        except AcceptInput as e:
                            return {curinput.input: e.args[0]}
                        except Exception as e:
                            vd.exceptionCaught(e)
                            break

                if r.keystroke == 'BUTTON1_RELEASED':
                    self.pressedLabel = None

                self.draw(self.scrForm)


@functools.wraps(VisiData.confirm)
@VisiData.api
def confirm(vd, prompt, exc=EscapeException):
    'Display *prompt* on status line and demand input that starts with "Y" or "y" to proceed. Return True when proceeding, otherwise raise *exc*, or if *exc* is falsy, return False'
    if vd.options.batch:
        return vd.fail('cannot confirm in batch mode: ' + prompt)

    form = FormSheet('confirm', rows=[
        AttrDict(x=2, y=0, text=' confirm ', color='yellow'),
        AttrDict(x=2, y=1, text=prompt, color='yellow'),
        AttrDict(x=.25, y=2, text=' yes ', color='black on yellow bold', key='y'),
        AttrDict(x=.75, y=2, text=' no ', color='black on yellow bold', key='n'),
        AttrDict(input='yn', keystrokes=['y', 'n', ENTER]),
    ])

    ret = FormCanvas(source=form).run(vd.scrFull)
    confirmed = False  # default is to disconfirm, if user exited the confirmation via: Esc ^C ^Q q
    if ret:
        yn = ret['yn'][:1]
        confirmed = yn and yn in 'Yy'
    if not confirmed:
        msg = 'disconfirmed: ' + prompt
        if exc:
            raise exc(msg)
        vd.warning(msg)
        return False
    return True



import curses
import functools
from copy import copy
from collections import namedtuple
from dataclasses import dataclass

from visidata import vd, options, Extensible, drawcache, drawcache_property, VisiData
import visidata

__all__ = ['ColorAttr', 'colors', 'update_attr', 'ColorMaker', 'rgb_to_attr']

vd.help_color = '''Color syntax: `<attribute> <fg-color> on <bg-color>`

- attributes: [:bold]bold[/] [:underline]underline[/] [:italic]italic[/] [:reverse]reverse[/]
- colors: 0-255 or [:black on 238]black[/] [:red on 238]red[/] [:green on 238]green[/] [:yellow on 238]yellow[/] [:blue on 238]blue[/] [:magenta on 238]magenta[/] [:cyan on 238]cyan[/] [:white on 238]white[/]
- the second color is used as a fallback if the first color is not available

See [:onclick https://visidata.org/docs/colors]https://visidata.org/docs/colors[/] for more detailed info.
'''


@dataclass
class ColorAttr:
    fg:int = -1         # default is no foreground specified
    bg:int = -1         # default is no background specified
    attributes:int = 0       # default is no attributes
    precedence:int = 0  # default is lowest priority
    colorname:str = ''

    def __init__(self, fg:int=-1, bg:int=-1, attributes:int=0, precedence:int=0, colorname:str=''):
        assert fg < 256, fg
        assert bg < 256, bg
        self.fg = fg
        self.bg = bg
        self.attributes = attributes
        self.precedence = precedence
        self.colorname = colorname

    def update(self, b:'ColorAttr', prec:int=10) -> 'ColorAttr':
        return update_attr(self, b, prec)

    @property
    def attr(self) -> int:
        a = colors._get_colorpair(self.fg, self.bg, self.colorname) | self.attributes
        assert a >= 0, a
        return a

    def as_str(self) -> str:
        attrnames = [attrname for attrname, attr in colors._attrs.items() if self.attributes & attr]
        attrnames.append(f'{self.fg} on {self.bg}')
        return ' '.join(attrnames)


def update_attr(oldattr:ColorAttr, updattr:ColorAttr, updprec:int=None) -> ColorAttr:
    assert isinstance(updattr, ColorAttr), updattr
    if updprec is None:
        updprec = updattr.precedence
    updfg = updattr.fg
    updbg = updattr.bg
    updattr = updattr.attributes

    # starting values, work backwards
    newfg = oldattr.fg
    newbg = oldattr.bg
    newattr = oldattr.attributes | updattr
    newprec = oldattr.precedence

    if newfg < 0 or (updfg >= 0 and updprec > newprec):
        newfg = updfg
    if newbg < 0 or (updbg >= 0 and updprec > newprec):
        newbg = updbg

    newprec = max(updprec, newprec)

    return ColorAttr(newfg, newbg, newattr, newprec)


class ColorMaker:
    def __init__(self):
        self.color_pairs = {}  # (fg,bg) -> (pairnum, colornamestr) (pairnum can be or'ed with other attrs)
        self.color_cache = {}  # colorname -> colorpair

    @drawcache_property
    def colorcache(self):
        return {}

    def setup(self):
        try:
            curses.use_default_colors()
        except Exception as e:
            pass

    @drawcache_property
    def colors(self):
        'not computed until curses color has been initialized'
        return {x[6:]:getattr(curses, x) for x in dir(curses) if x.startswith('COLOR_') and x != 'COLOR_PAIRS'}

    def __getitem__(self, colorname:str) -> ColorAttr:
        'colors["green"] or colors["foo"] returns parsed ColorAttr.'
        return self.get_color(colorname)

    def __getattr__(self, optname) -> ColorAttr:
        'colors.color_foo or colors.foo returns parsed ColorAttr.'
        return self.get_color(optname)

    @drawcache
    def resolve_colors(self, colorstack):
        'Returns the ColorAttr for the colorstack, a list of (prec, color_option_name) sorted highest-precedence color first.'
        cattr = ColorAttr()
        for prec, coloropt in colorstack:
            c = self.get_color(coloropt)
            cattr = update_attr(cattr, c, prec)
        return cattr

    def get_color(self, optname:str, precedence:int=0) -> ColorAttr:
        '''Return ColorAttr for options.color_foo if *optname* of either "foo" or "color_foo",
           Otherwise parse *optname* for colorstring like "bold 34 red on 135 blue".'''
        r = self.colorcache.get(optname, None)
        if r is None:
            coloropt = vd.options._get(optname) or vd.options._get(f'color_{optname}')
            colornamestr = coloropt.value if coloropt else optname
            r = self.colorcache[optname] = self._colornames_to_cattr(colornamestr, precedence)
        return r

    def _split_colorstr(self, colorstr):
        'Return (fgstr, bgstr, attrlist) parsed from colorstr.'
        fgbgattrs = ['', '', []]  # fgstr, bgstr, attrlist
        if not colorstr:
            return fgbgattrs
        colorstr = str(colorstr)

        i = 0  # fg by default
        for x in colorstr.split():
            if x == 'fg':
                i = 0
                continue
            elif x in ['on', 'bg']:
                i = 1
                continue

            if hasattr(curses, 'A_' + x.upper()):
                fgbgattrs[2].append(x)
            else:
                if not fgbgattrs[i]:  # keep first known color
                    if self._get_colornum(x) is not None:   # only set known colors
                        fgbgattrs[i] = x
                    else:
                        fgbgattrs[i] = None

        return fgbgattrs

    def _get_colornum(self, colorname:'str|int', default:int=-1) -> int:
        'Return terminal color number for colorname.'
        if isinstance(colorname, int):
            return colorname

        if not colorname:
            return default

        r = self.color_cache.get(colorname, None)
        if r is not None:
            return r

        try:
            r = int(colorname)
        except Exception:
            r = self.colors.get(colorname.upper())

        if r is None:
            return None

        try: # test to see if color is available
            curses.init_pair(255, r, 0)
            self.color_cache[colorname] = r
            return r
        except curses.error as e:
            return None  # not available
        except ValueError:  # Python 3.10+  issue #1227
            return None

    def _attrnames_to_num(self, attrnames:'list[str]') -> int:
        attrs = 0
        for attr in attrnames:
            attrs |= getattr(curses, 'A_'+attr.upper())
        return attrs

    @drawcache_property
    def _attrs(self):
        return {k[2:].lower():getattr(curses, k) for k in dir(curses) if k.startswith('A_') and k != 'A_ATTRIBUTES'}

    @drawcache
    def _colornames_to_cattr(self, colorname:str, precedence=0) -> ColorAttr:
        fg, bg, attrlist = self._split_colorstr(colorname)

        fg = self._get_colornum(fg)
        bg = self._get_colornum(bg)
        return ColorAttr(fg, bg,
                         self._attrnames_to_num(attrlist),
                         precedence, colorname)

    def _get_colorpair(self, fg:'int|None', bg:'int|None', colorname:str) -> int:
            pairnum, _ = self.color_pairs.get((fg, bg), (None, ''))
            if pairnum is None:
                if len(self.color_pairs) > 254:
                    self.color_pairs.clear()  # start over
                    self.color_cache.clear()
                pairnum = len(self.color_pairs)+1
                if fg is None: fg = -1
                if bg is None: bg = -1
                try:
                    curses.init_pair(pairnum, fg, bg)
                except curses.error as e:
                    return 0  # do not cache
                self.color_pairs[(fg, bg)] = (pairnum, colorname)

            return curses.color_pair(pairnum)


colors = ColorMaker()

def rgb_to_xterm256(r:int,g:int,b:int,a:int=255) -> int:
    if a == 0:
        return -1

    if max(r,g,b) - min(r,g,b) < 8:
        if r <= 4: return 16
        elif r <= 8: return 232
        elif r >= 247: return 231
        elif r >= 238: return 255
        else:
            return int(232 + (r-8)//10)
    else:
        r = max(0, r-(95-40)) // 40
        g = max(0, g-(95-40)) // 40
        b = max(0, b-(95-40)) // 40
        return int(16 + r*36 + g*6 + b)


@functools.lru_cache(256)
def rgb_to_attr(r:int,g:int,b:int,a:int=255) -> str:
    return str(rgb_to_xterm256(r,g,b,a))


import sys
vd.addGlobals({k:getattr(sys.modules[__name__], k) for k in __all__})



from visidata import VisiData, vd, Sheet, date, anytype, options, Column, ItemColumn, Progress, vlen, PyobjSheet, TypedWrapper
from itertools import chain

'Loaders for .npy and .npz.  Save to .npy.  Depends on the zip loader.'

@VisiData.api
def open_npy(vd, p):
    return NpySheet(p.base_stem, source=p)

@VisiData.api
def open_npz(vd, p):
    return NpzSheet(p.base_stem, source=p)

vd.option('npy_allow_pickle', False, 'numpy allow unpickling objects (unsafe)')
vd.option('npy_matrix_enumerate', False, 'enumerate matrix rows and columns')

class NpySheet(Sheet):
    def iterload(self):
        numpy = vd.importExternal('numpy')
        if not hasattr(self, 'npy'):
            self.npy = numpy.load(str(self.source), encoding='bytes', allow_pickle=bool(self.options.npy_allow_pickle))
        self.reloadCols()
        transpose = len(self.npy.shape)==1 and not bool(self.npy.dtype.names)
        if transpose:
            source = self.npy[:,None]
        else:
            source = self.npy

        nrows = len(self.npy)

        if self.options.npy_matrix_enumerate:
            source = list(list((chain((i,), row))) for i, row in enumerate(source))

        yield from Progress(source, nrows)


    def reloadCols(self):
        self.columns = []
        if len(self.npy.shape)==1:
            for i, (colname, fmt, *shape) in enumerate(self.npy.dtype.descr):
                if not colname:
                    colname = f"col{i}"
                ctype = _guess_type(shape, fmt)
                if ctype=="time":
                    self.addColumn(Column(colname, type=date, getter=lambda c,r,i=i: str(r[i])))
                    continue
                self.addColumn(ItemColumn(colname, i, type=ctype))
        elif len(self.npy.shape)==2:
            ncols = self.npy.shape[1]
            ctype = _guess_type(None, self.npy.dtype.descr[0][1])

            if self.options.npy_matrix_enumerate:
                self.addColumn(ItemColumn("row", 0, width=8, keycol=1, type=int), index=0)
                for i in range(ncols):
                    self.addColumn(ItemColumn(f'col{i}', i+1, width=8, type=ctype), index=i+1)
            else:
                for i in range(ncols):
                    self.addColumn(ItemColumn('', i, width=8, type=ctype), index=i)
        else:
            vd.fail(f"too many dimensions in shape {self.npy.shape}")

def _guess_type(shape, fmt):
    if shape:
        return anytype
    elif 'M' in fmt:
        return "time"
    elif 'i' in fmt or 'u' in fmt:
        return int
    elif 'f' in fmt:
        return float
    return anytype

class NpzSheet(vd.ZipSheet):
    # rowdef: tuple(tablename, table)
    columns = [
        ItemColumn('name', 0),
        ItemColumn('length', 1, type=vlen),
    ]

    def iterload(self):
        numpy = vd.importExternal('numpy')
        self.npz = numpy.load(str(self.source), encoding='bytes', allow_pickle=bool(self.options.npy_allow_pickle))
        yield from Progress(self.npz.items())

    def openRow(self, row):
        numpy = vd.importExternal('numpy')
        tablename, tbl = row
        if isinstance(tbl, numpy.ndarray):
            return NpySheet(tablename, npy=tbl)

        return PyobjSheet(tablename, source=tbl)


@VisiData.api
def save_npy(vd, p, sheet):
    np = vd.importExternal('numpy')

    dtype = []

    for col in Progress(sheet.visibleCols):
        if col.type in (int, vlen):
            dt = 'i8'
        elif col.type is date:
            dt = 'datetime64[s]'
        elif col.type in vd.numericTypes:
            dt = 'f8'
        else: #  if col.type in (str, anytype):
            width = col.getMaxDataWidth(sheet.rows)
            dt = 'U'+str(width)
        dtype.append((col.name, dt))

    data = []
    for typedvals in sheet.iterdispvals(format=False):
        nprow = []
        for col, val in typedvals.items():
            if isinstance(val, TypedWrapper):
                if col.type is anytype:
                    val = ''
                else:
                    val = options.safe_error
            elif col.type is date:
                val = np.datetime64(val.isoformat())
            nprow.append(val)
        data.append(tuple(nprow))

    arr = np.array(data, dtype=dtype)
    with p.open_bytes(mode='w') as outf:
        np.save(outf, arr, allow_pickle=bool(sheet.options.npy_allow_pickle))



# SPDX-License-Identifier: MIT

from setuptools import setup, find_packages
from pathlib import Path


exec(Path('__about__.py').read_text())


def readme():
    return Path("README.md").read_text()


def requirements():
    return Path("requirements.txt").read_text().splitlines()

def requirements_extra():
    return Path("requirements-extra.txt").read_text().splitlines()


setup(
        name="vdsql",
        version=__version__,
        description=__description__,
        long_description=readme(),
        long_description_content_type="text/markdown",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Programming Language :: Python :: 3",
        ],
        keywords="visidata sql rdbms ibis substrait",
        author="Saul Pwanson",
        url="https://github.com/visidata/vdsql",
        python_requires=">=3.9",
        packages=find_packages(exclude=["tests"]),
        entry_points={'visidata.plugins': 'vdsql=visidata.apps.vdsql'},
        scripts=['vdsql'],
        install_requires=requirements(),
        extra_requires=requirements_extra(),
)



import time

from visidata import vd, VisiData

from ._ibis import IbisTableSheet, IbisConnectionPool, IbisTableIndexSheet


@VisiData.api
def openurl_snowflake(vd, p, filetype=None):
    return IbisTableIndexSheet(p.base_stem, source=p, filetype=None, database_name=None,
                               ibis_conpool=IbisConnectionPool(p),
                               sheet_type=SnowflakeSheet)


class SnowflakeSheet(IbisTableSheet):
    @property
    def countRows(self):
        r = super().countRows
        if r is None and self.cursor is None:
            return None  # no cursor yet
        return r

    def executeSql(self, sql):
        assert self.cursor is None

        with self.con as con:
            con = con.con

            if self.warehouse:
                con.execute(f'USE WAREHOUSE {self.warehouse}')

            with con.begin() as c:
                snowflake_conn = c.connection.dbapi_connection
                cursor = self.cursor = snowflake_conn.cursor()
                cursor.execute_async(sql)
                while snowflake_conn.is_still_running(snowflake_conn.get_query_status(cursor.sfqid)):
                    time.sleep(.1)

        cursor.get_results_from_sfqid(cursor.sfqid)
        yield from cursor.fetchall()

        self.cursor = None

    def iterload(self):
        try:
            with self.con as con:
                if self.query is None:
                    self.query = self.baseQuery(con)
                yield from self.executeSql(self.ibis_to_sql(self.withRowcount(self.baseQuery(con))))
        except BaseException:
            if self.cursor:
                self.cancelQuery(self.cursor.sfqid)
            raise

        self.reloadColumns(self.query)  # columns based on query without metadata

    def cancelQuery(self, qid):
        vd.status(f'canceling "{qid}"')
        with self.con as con:
            with con.begin() as con:
                cursor = con.connection.dbapi_connection.cursor()
                cursor.execute(f"SELECT SYSTEM$CANCEL_QUERY('{qid}')")
                vd.status(cursor.fetchall())


SnowflakeSheet.init('cursor', lambda: None)
SnowflakeSheet.init('warehouse', str)



'''
Marking selected rows with a keystroke, selecting marked rows,
and viewing lists of marks and their rows.
'''

from copy import copy
from visidata import vd, asyncthread, vlen, VisiData, TableSheet, ColumnItem, RowColorizer

@VisiData.lazy_property
def marks(vd):
    return MarksSheet('marks')


class MarkSheet(TableSheet):
    pass


class MarksSheet(TableSheet):
    '''
    The Marks Sheet shows all marks in use (on all sheets) and how many rows have each mark.
    '''
    rowtype = "marks"  # rowdef: [mark, color, [rows]]
    columns = [
        ColumnItem('mark', 0),
        ColumnItem('color', 1),
        ColumnItem('rows', 2, type=vlen),
    ]
    colorizers = [
        RowColorizer(2, None, lambda s,c,r,v: r and r[1])
    ]
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.marknotes = list('0123456789')
        self.marks = []  #
        self.markedRows = {}  # rowid(row): [row, set(marks)]
        self.rows = []

    def getColor(self, sheet, row):
        mark = self.getMark(sheet, row)
        if not mark:
            return ''
        return self.getMarkRow(sheet, mark)[1]

    def getMark(self, sheet, row):
        mrow = self.markedRows.get(sheet.rowid(row), None)
        if not mrow:
            return ''
        if mrow[1]:
            return next(iter(mrow[1]))  # first item in set

    def getMarks(self, row):
        'Return set of all marks for given row'
        return self.markedRows[self.rowid(row)][1]

    def isMarked(self, row, mark):
        'Return True if given row has given mark'
        return mark in self.getMarks(row)

    def getMarkRow(self, sheet, mark):
        for r in self.rows:
            if r[0] == mark:
                return r
        r = [mark, 'color_note_type', MarkSheet('mark_', rows=[], columns=copy(sheet.columns))]
        self.addRow(r)
        return r

    def setMark(self, sheet, row, mark):
        rowid = self.rowid(row)
        if rowid not in self.markedRows:
            self.markedRows[rowid] = [row, set(mark)]
        else:
            self.markedRows[rowid][1].add(mark)

        vd.marks.getMarkRow(sheet, mark)[2].addRow(row)

    def unsetMark(self, sheet, row, mark):
        rowid = self.rowid(row)
        if rowid in self.markedRows:
            self.markedRows[rowid][1].remove(mark)
        vd.marks.getMarkRow(sheet, mark)[2].deleteBy(lambda r,x=row: r is x)

    def inputmark(self):
        return vd.inputsingle('mark: ') or self.marknotes.pop(0)

    def openRow(self, row):
        return row[2]


@VisiData.api
@asyncthread
def mark(vd, sheet, rows, m):
    for r in rows:
        vd.marks.setMark(sheet, r, m)

@VisiData.api
@asyncthread
def unmark(vd, sheet, rows, m):
    for r in rows:
        vd.marks.unsetMark(sheet, r, m)


vd.rowNoters.insert(0, lambda sheet, row: vd.marks.getMark(sheet, row))

TableSheet.colorizers.append(RowColorizer(2, None, lambda s,c,r,v: not c and r and vd.marks.getColor(s, r)))

TableSheet.addCommand('', 'mark-row', 'vd.mark(sheet, [cursorRow], vd.marks.inputmark())', '')
TableSheet.addCommand('', 'unmark-row', 'vd.unmark(sheet, [cursorRow], vd.marks.inputmark())', '')
TableSheet.addCommand('', 'mark-selected', 'vd.mark(sheet, selectedRows, vd.marks.inputmark())', '')
TableSheet.addCommand('', 'unmark-selected', 'vd.unmark(sheet, selectedRows, vd.marks.inputmark())', '')

TableSheet.addCommand('', 'select-marks', 'select(gatherBy(lambda r,mark=vd.marks.inputmark(): vd.marks.isMarked(r, mark)), progress=False)', '')
TableSheet.addCommand('', 'stoggle-marks', 'toggle(gatherBy(lambda r,mark=vd.marks.inputmark(): vd.marks.isMarked(r, mark)), progress=False)', '')
TableSheet.addCommand('', 'unselect-marks', 'unselect(gatherBy(lambda r,mark=vd.marks.inputmark(): vd.marks.isMarked(r, mark)), progress=False)', '')

TableSheet.addCommand('', 'open-marks', 'vd.push(vd.marks)', '')

TableSheet.addCommand('', 'go-prev-mark', 'moveToNextRow(lambda row,mark=vd.marks.inputmark(): vd.marks.isMarked(row, mark), reverse=True, msg="no previous marked row")', 'go up current column to previous row with given mark')
TableSheet.addCommand('', 'go-next-mark', 'moveToNextRow(lambda row,mark=vd.marks.inputmark(): vd.marks.isMarked(row, mark), msg="no next marked row")', 'go down current column to next row with given mark')


vd.addMenuItems('''
    View > Marks > open-marks
    Row > Mark > open Marks Sheet > open-marks
    Row > Mark > current row > mark-row
    Row > Mark > selected rows > mark-selected
    Row > Unmark > current row > unmark-row
    Row > Unmark > selected rows > unmark-selected
    Row > Select > marked rows > select-marks
    Row > Unselect > marked rows > unselect-marks
    Row > Toggle select > marked rows > stoggle-marks
    Row > Goto > next marked row > go-next-mark
    Row > Goto > previous marked row > go-prev-mark
''')



'''
Specify the billing_project_id as the netloc, and the actual dataset_id as the path:

    vdsql bigquery://<billing_project>/<dataset_id>''
'''


from visidata import vd, VisiData, Sheet, AttrColumn
from . import IbisTableSheet, IbisTableIndexSheet, IbisConnectionPool

import ibis
import ibis.expr.operations as ops

@VisiData.api
def openurl_bigquery(vd, p, filetype=None):
    vd.configure_ibis()
    vd.configure_bigquery()
    return BigqueryDatabaseIndexSheet(p.base_stem, source=p, ibis_con=None)

vd.openurl_bq = vd.openurl_bigquery


@VisiData.api
def configure_bigquery(vd):
    @ibis.bigquery.add_operation(ops.TimestampDiff)
    def bq_timestamp_diff(t, expr):
        op = expr.op()
        left = t.translate(op.left)
        right = t.translate(op.right)
        return f"TIMESTAMP_DIFF({left}, {right}, SECOND)"


class BigqueryDatabaseIndexSheet(Sheet):
    rowtype = 'databases'  # rowdef: DatasetListItem
    columns = [
#        AttrColumn('project', width=0),
        AttrColumn('dataset_id'),
        AttrColumn('friendly_name'),
        AttrColumn('full_dataset_id', width=0),
        AttrColumn('labels'),
    ]
    nKeys = 1

    @property
    def con(self):
        if not self.ibis_con:
            import ibis
            self.ibis_con = ibis.connect(self.source)
        return self.ibis_con

    def iterload(self):
        yield from self.con.client.list_datasets(project=self.source.name)

    def openRow(self, row):
        return IbisTableIndexSheet(row.dataset_id,
                                   database_name=self.source.name+'.'+row.dataset_id,
                                   ibis_con=self.con,
                                   ibis_conpool=IbisConnectionPool(f"{self.source}/{row.dataset_id}"),
                                   source=row,
                                   filetype=None,
                                   sheet_type=IbisTableSheet)



import re
from visidata import vd, VisiData, Sheet, IndexSheet, SequenceSheet, ColumnItem, Path, AttrDict, ColumnAttr, asyncthread

SPREADSHEET_FIELDS='properties sheets namedRanges spreadsheetUrl developerMetadata dataSources dataSourceSchedules'.split()
SHEET_FIELDS='merges conditionalFormats filterViews protectedRanges basicFilter charts bandedRanges developerMetadata rowGroups columnGroups slicers'.split()

@VisiData.api
def open_gsheets(vd, p):
    m = re.search(r'([A-z0-9_]{44})', p.given)
    if m:
        return GSheetsIndex(p.base_stem, source=m.groups()[0])

vd.open_g = vd.open_gsheets

@VisiData.lazy_property
def google_discovery(self):
    googleapiclient = vd.importExternal('googleapiclient', 'google-api-python-client')
    from googleapiclient import discovery
    return discovery


@VisiData.cached_property
def _gsheets(vd):
    return vd.google_discovery.build("sheets", "v4", credentials=vd.google_auth('spreadsheets.readonly')).spreadsheets()


@VisiData.cached_property
def _gsheets_rw(vd):
    return vd.google_discovery.build("sheets", "v4", credentials=vd.google_auth('spreadsheets')).spreadsheets()


class GSheetsIndex(Sheet):
    columns = [
        ColumnAttr('title', 'properties.title'),
        ColumnAttr('type', 'properties.sheetType', width=0),
        ColumnAttr('nRows', 'properties.gridProperties.rowCount', type=int),
        ColumnAttr('nCols', 'properties.gridProperties.columnCount', type=int),
    ]
    def iterload(self):
        googlesheet = vd._gsheets.get(spreadsheetId=self.source, fields=','.join(SPREADSHEET_FIELDS)).execute()
        vd.status(googlesheet['properties']['title'])

        for gsheet in googlesheet['sheets']:
            yield AttrDict(gsheet)

    def openRow(self, r):
        return GSheet(r.properties.title, source=self.source)


class GSheet(SequenceSheet):
    '.source is gsheet id; .name is sheet name'
    def iterload(self):
        result = vd._gsheets.values().get(spreadsheetId=self.source, range=self.name).execute()
        yield from result.get('values', [])


@VisiData.api
def save_gsheets(vd, p, *sheets):
    gsheet = vd._gsheets_rw.create(body={
        'properties': { 'title': p.base_stem },
        'sheets': list({'properties': { 'title': vs.name }} for vs in sheets),
    }, fields='spreadsheetId').execute()

    gsheetId = gsheet.get('spreadsheetId')
    vd.status(f'https://docs.google.com/spreadsheets/d/{gsheetId}/')

    for vs in sheets:
        rows = [list(c.name for c in vs.visibleCols)]
        rows += list(list(val for col, val in row.items())
                        for row in vs.iterdispvals(*vs.visibleCols, format=True))

        vd._gsheets_rw.values().append(
            spreadsheetId=gsheetId,
            valueInputOption='RAW',
            range=vs.name,
            body=dict(values=rows)
        ).execute()

vd.save_g = vd.save_gsheets



import itertools

from visidata import vd, BaseSheet, Sheet, Column, Progress, ALT, asyncthread


def rotateRange(n, idx, reverse=False):
    'Wraps an iter starting from idx. Yields indices from idx+1 to n and then 0 to idx, or from idx-1 to 0 and n-1 to idx.'
    if n == 0: return []
    if reverse:
        rng = range(idx-1, -1, -1)
        rng2 = range(n-1, idx-1, -1)
    else:
        rng = range(idx+1, n)
        rng2 = range(0, idx+1)

    wrapped = False
    with Progress(total=n) as prog:
        for r in itertools.chain(rng, rng2):
            prog.addProgress(1)
            if not wrapped and r in rng2:
                vd.status('search wrapped')
                wrapped = True
            yield r

@Sheet.api
def pageLeft(self):
    '''Redraw page one screen to the left.

    Note: keep the column cursor in the same general relative position:

     - if it is on the furthest right column, then it should stay on the
       furthest right column if possible

     - likewise on the left or in the middle

    So really both the `leftIndex` and the `cursorIndex` should move in
    tandem until things are correct.'''

    targetIdx = self.leftVisibleColIndex  # for rightmost column
    firstNonKeyVisibleColIndex = self.visibleCols.index(self.nonKeyVisibleCols[0])
    while self.rightVisibleColIndex != targetIdx and self.leftVisibleColIndex > firstNonKeyVisibleColIndex:
        self.cursorVisibleColIndex -= 1
        self.leftVisibleColIndex -= 1
        self.calcColLayout()  # recompute rightVisibleColIndex

    # in case that rightmost column is last column, try to squeeze maximum real estate from screen
    if self.rightVisibleColIndex == self.nVisibleCols-1:
        # try to move further left while right column is still full width
        while self.leftVisibleColIndex > 0:
            rightcol = self.visibleCols[self.rightVisibleColIndex]
            if (rightcol.width or 0) > self._visibleColLayout[self.rightVisibleColIndex][1]:
                # went too far
                self.cursorVisibleColIndex += 1
                self.leftVisibleColIndex += 1
                break
            else:
                self.cursorVisibleColIndex -= 1
                self.leftVisibleColIndex -= 1
                self.calcColLayout()  # recompute rightVisibleColIndex


@Sheet.api
@asyncthread
def moveToNextRow(vs, func, reverse=False, msg='no different value up this column'):
    'Move cursor to next (prev if reverse) row for which func returns True.  Returns False if no row meets the criteria.'
    rng = range(vs.cursorRowIndex-1, -1, -1) if reverse else range(vs.cursorRowIndex+1, vs.nRows)
    found = False
    with Progress(total=len(vs.rows)) as prog:
        for i in rng:
            prog.addProgress(1)
            try:
                if func(vs.rows[i]):
                    vs.cursorRowIndex = i
                    found = True
                    break
            except Exception:
                pass

    if not found:
        vd.status(msg)


@Column.property
def visibleWidth(self):
    'Width of column as is displayed in terminal'
    vcolidx = self.sheet.visibleCols.index(self)
    if vcolidx in self.sheet._visibleColLayout:
        w = self.sheet._visibleColLayout[vcolidx][1]
    else:  #this case should never happen in normal use
        #the width can be inaccurate if the column is not at x=0
        w = self.sheet.calcSingleColLayout(vcolidx)
    return w


Sheet.addCommand(None, 'go-left',  'cursorRight(-1)', 'go left', replay=False)
Sheet.addCommand(None, 'go-down',  'cursorDown(+1)', 'go down', replay=False)
Sheet.addCommand(None, 'go-up',    'cursorDown(-1)', 'go up', replay=False)
Sheet.addCommand(None, 'go-right', 'cursorRight(+1)', 'go right', replay=False)
Sheet.addCommand(None, 'go-pagedown', 'cursorDown(nScreenRows-1); sheet.topRowIndex = bottomRowIndex', 'scroll one page forward', replay=False)
Sheet.addCommand(None, 'go-pageup', 'cursorDown(-nScreenRows+1); sheet.bottomRowIndex = topRowIndex', 'scroll one page backward', replay=False)

Sheet.addCommand(None, 'go-leftmost', 'sheet.cursorVisibleColIndex = sheet.leftVisibleColIndex = 0', 'go all the way to the left of sheet')
Sheet.addCommand(None, 'go-top', 'sheet.cursorRowIndex = sheet.topRowIndex = 0', 'go all the way to the top of sheet')
Sheet.addCommand(None, 'go-bottom', 'sheet.cursorRowIndex = sheet.bottomRowIndex = len(rows)-1', 'go all the way to the bottom of sheet')
Sheet.addCommand(None, 'go-rightmost', 'sheet.leftVisibleColIndex = len(visibleCols)-1; pageLeft(); sheet.cursorVisibleColIndex = len(visibleCols)-1', 'go all the way to the right of sheet')

Sheet.addCommand('zr', 'go-row-number', 'sheet.cursorRowIndex = int(input("move to row number: "))', 'go to the given row number (0-based)')


Sheet.addCommand('<', 'go-prev-value', 'moveToNextRow(lambda row,sheet=sheet,col=cursorCol,val=cursorTypedValue: col.getTypedValue(row) != val, reverse=True, msg="no different value up this column")', 'go up current column to next value')
Sheet.addCommand('>', 'go-next-value', 'moveToNextRow(lambda row,sheet=sheet,col=cursorCol,val=cursorTypedValue: col.getTypedValue(row) != val, msg="no different value down this column")', 'go down current column to next value')
Sheet.addCommand('{', 'go-prev-selected', 'moveToNextRow(lambda row,sheet=sheet: sheet.isSelected(row), reverse=True, msg="no previous selected row")', 'go up current column to previous selected row')
Sheet.addCommand('}', 'go-next-selected', 'moveToNextRow(lambda row,sheet=sheet: sheet.isSelected(row), msg="no next selected row") ', 'go down current column to next selected row')

Sheet.addCommand('z<', 'go-prev-null', 'moveToNextRow(lambda row,col=cursorCol,isnull=isNullFunc(): isnull(col.getValue(row)), reverse=True, msg="no null up this column")', 'go up current column to next null value')
Sheet.addCommand('z>', 'go-next-null', 'moveToNextRow(lambda row,col=cursorCol,isnull=isNullFunc(): isnull(col.getValue(row)), msg="no null down this column")', 'go down current column to next null value')

for i in range(1, 11):
    BaseSheet.addCommand(ALT+str(i)[-1], 'jump-sheet-'+str(i), f'vd.push(*(list(s for s in allSheets if s.shortcut==str({i})) or fail("no sheet")))', f'jump to sheet {i}')

for i in range(11, 21):
    BaseSheet.addCommand('', 'jump-sheet-'+str(i), f'vd.push(*(list(s for s in allSheets if s.shortcut==str({i})) or fail("no sheet")))', f'jump to sheet {i}')

BaseSheet.bindkey('KEY_LEFT', 'go-left')
BaseSheet.bindkey('KEY_DOWN', 'go-down')
BaseSheet.bindkey('KEY_UP', 'go-up')
BaseSheet.bindkey('KEY_RIGHT', 'go-right')
BaseSheet.bindkey('KEY_NPAGE', 'go-pagedown')
BaseSheet.bindkey('KEY_PPAGE', 'go-pageup')

BaseSheet.bindkey('gKEY_LEFT', 'go-leftmost'),
BaseSheet.bindkey('gKEY_RIGHT', 'go-rightmost'),
BaseSheet.bindkey('gKEY_UP', 'go-top'),
BaseSheet.bindkey('gKEY_DOWN', 'go-bottom'),
BaseSheet.bindkey('Home', 'go-top')
BaseSheet.bindkey('End', 'go-bottom')

Sheet.bindkey('BUTTON1_CLICKED', 'go-mouse')
Sheet.bindkey('BUTTON3_PRESSED', 'go-mouse')

# vim-style scrolling with the 'z' prefix
Sheet.addCommand('zz', 'scroll-middle', 'sheet.topRowIndex = cursorRowIndex-int(nScreenRows/2)', 'scroll current row to center of screen')

Sheet.addCommand('kRIT5', 'go-right-page', 'sheet.cursorVisibleColIndex = sheet.leftVisibleColIndex = rightVisibleColIndex', 'scroll cursor one page right', replay=False)
Sheet.addCommand('kLFT5', 'go-left-page', 'pageLeft()', 'scroll cursor one page left', replay=False)
Sheet.addCommand(None, 'scroll-left', 'sheet.cursorVisibleColIndex -= options.scroll_incr', 'scroll one column left')
Sheet.addCommand(None, 'scroll-right', 'sheet.cursorVisibleColIndex += options.scroll_incr', 'scroll one column right')
Sheet.addCommand(None, 'scroll-leftmost', 'sheet.leftVisibleColIndex = cursorVisibleColIndex', 'scroll sheet to leftmost column')
Sheet.addCommand(None, 'scroll-rightmost', 'tmp = cursorVisibleColIndex; pageLeft(); sheet.cursorVisibleColIndex = tmp', 'scroll sheet to rightmost column')

Sheet.addCommand('zl', 'scroll-cells-right', 'cursorCol.hoffset += cursorCol.visibleWidth-2', 'scroll display of current column to the right')
Sheet.addCommand('zh', 'scroll-cells-left', 'cursorCol.hoffset -= cursorCol.visibleWidth-2', 'scroll display of current column to the left')
Sheet.addCommand('gzl', 'scroll-cells-rightmost', 'cursorCol.hoffset = -cursorCol.visibleWidth+2', 'scroll display of current column to the end')
Sheet.addCommand('gzh', 'scroll-cells-leftmost', 'cursorCol.hoffset = 0', 'scroll display of current column to the beginning')

Sheet.addCommand('zj', 'scroll-cells-down', 'cursorCol.voffset += 1 if cursorCol.height > 1 else fail("multiline column needed for scrolling")', 'scroll display of current column down one line')
Sheet.addCommand('zk', 'scroll-cells-up', 'cursorCol.voffset -= 1 if cursorCol.height > 1 else fail("multiline column needed for scrolling")', 'scroll display of current column up one line')
Sheet.addCommand('gzj', 'scroll-cells-bottom', 'cursorCol.voffset = -1', 'scroll display of current column to the bottom')
Sheet.addCommand('gzk', 'scroll-cells-top', 'cursorCol.voffset = 0', 'scroll display of current column to the top')

Sheet.addCommand(None, 'go-end',  'sheet.cursorRowIndex = len(rows)-1; sheet.cursorVisibleColIndex = len(visibleCols)-1', 'go to last row and last column')
Sheet.addCommand(None, 'go-home', 'sheet.topRowIndex = sheet.cursorRowIndex = 0; sheet.leftVisibleColIndex = sheet.cursorVisibleColIndex = 0', 'go to first row and first column')

BaseSheet.bindkey('Ctrl+ScrollUp', 'scroll-left')
BaseSheet.bindkey('Ctrl+ScrollDown', 'scroll-right')

BaseSheet.bindkey('zKEY_UP', 'scroll-up')
BaseSheet.bindkey('zKEY_DOWN', 'scroll-down')
BaseSheet.bindkey('zKEY_LEFT', 'scroll-left')
BaseSheet.bindkey('zKEY_RIGHT', 'scroll-right')

# vim-like keybindings

BaseSheet.bindkey('h', 'go-left'),
BaseSheet.bindkey('j', 'go-down'),
BaseSheet.bindkey('k', 'go-up'),
BaseSheet.bindkey('l', 'go-right'),
BaseSheet.bindkey('^F', 'go-pagedown'),
BaseSheet.bindkey('^B', 'go-pageup'),
BaseSheet.bindkey('gg', 'go-top'),
BaseSheet.bindkey('G',  'go-bottom'),
BaseSheet.bindkey('gj', 'go-bottom'),
BaseSheet.bindkey('gk', 'go-top'),
BaseSheet.bindkey('gh', 'go-leftmost'),
BaseSheet.bindkey('gl', 'go-rightmost')

BaseSheet.addCommand('^^', 'jump-prev', 'vd.activeStack[1:] or fail("no previous sheet"); vd.push(vd.activeStack[1])', 'jump to previous sheet in this pane')
BaseSheet.addCommand('g^^', 'jump-first', 'vd.push(vd.activeStack[-1])', 'jump to first sheet')

BaseSheet.addCommand('BUTTON1_RELEASED', 'no-op', 'pass', 'do nothing')

BaseSheet.addCommand(None, 'mouse-enable', 'mm, _ = curses.mousemask(-1); status("mouse "+("ON" if mm else "OFF"))', 'enable mouse events')
BaseSheet.addCommand(None, 'mouse-disable', 'mm, _ = curses.mousemask(0); status("mouse "+("ON" if mm else "OFF"))', 'disable mouse events')


vd.addGlobals({'rotateRange': rotateRange})

vd.addMenuItems('''
    View > Other sheet > previous sheet > jump-prev
    View > Other sheet > first sheet > jump-first
    Row > Goto > top > go-top
    Row > Goto > bottom > go-bottom
    Row > Goto > previous > page > go-pageup
    Row > Goto > previous > null > go-prev-null
    Row > Goto > previous > value > go-prev-value
    Row > Goto > previous > selected > go-prev-selected
    Row > Goto > next > page > go-pagedown
    Row > Goto > next > null > go-next-null
    Row > Goto > next > value > go-next-value
    Row > Goto > next > selected > go-next-selected
    Row > Goto > by number > go-row-number
    View > Other sheet > previous sheet > jump-prev
    View > Other sheet > first sheet > jump-first
''')



from visidata import vd, VisiData, JsonSheet


@VisiData.api
def open_msgpack(vd, p):
    return MsgpackSheet(p.name, source=p)


VisiData.open_msgpackz = VisiData.open_msgpack


class MsgpackSheet(JsonSheet):
    def iterload(self):
        msgpack = vd.importModule('msgpack')
        data = self.source.read_bytes()
        if self.options.filetype == 'msgpackz':
            brotli = vd.importModule('brotli')
            data = brotli.decompress(data)
        yield from msgpack.unpackb(data, raw=False)



from visidata import vd, Sheet

Sheet.addCommand(None, 'random-rows', 'nrows=int(input("random number to filter: ", value=nRows)); vs=copy(sheet); vs.name=name+"_sample"; vs.rows=sample(rows, nrows or nRows); vd.push(vs)', 'open duplicate sheet with a random population subset of N rows')
Sheet.addCommand(None, 'select-random', 'nrows=int(input("random number to select: ", value=nRows)); select(sample(rows, nrows or nRows))', 'select random sample of N rows')

vd.addMenuItems('Row > Select > random sample > select-random')



__all__ = '__title__ __author__ __version__ __description__ __license__ __copyright__'.split()

__title__ = 'vdsql'
__author__ = 'Saul Pwanson <code@saul.pw>'
__version__ = '0.3dev'
__description__ = 'VisiData for database queries'
__license__ = 'Apache License, Version 2.0'
__copyright__ = 'Copyright 2022 ' + __author__



#!/usr/bin/env python3

def main():
    import ibis
    import visidata
    from visidata import main, vd

    from . import __version__
    visidata.__version_info__ = f'vdsql {__version__}'

    for ext in "db ddb duckdb sqlite sqlite3".split():
        setattr(vd, f"open_{ext}", vd.open_vdsql)

    for entry_point in ibis.util.backend_entry_points():
        if entry_point.name in ['bigquery', 'clickhouse', 'snowflake']:
            # these have their own custom openurl_ funcs already installed
            continue

        attrname = f"openurl_{entry_point.name}"
        # when running vdsql directly, override visidata builtin loader with vdsql loader #1929
        setattr(vd, attrname, vd.open_vdsql)

    main.vd_cli()


if __name__ == "__main__":
    main()



from visidata import vd, IndexSheet


vd.option('load_lazy', False, 'load subsheets always (False) or lazily (True)')
vd.option('skip', 0, 'skip N rows before header', replay=True)
vd.option('header', 1, 'parse first N rows as column names', replay=True)

IndexSheet.options.header = 0
IndexSheet.options.skip = 0



from visidata import Sheet, Column, DisplayWrapper


@Column.api
def displayer_url(self, dw:DisplayWrapper, width=None):
    'Display cell text as clickable url'
    yield ('onclick '+dw.text, dw.text)


Sheet.addCommand("", "type-url", "sheet.cursorCol.displayer = 'url'", "set column to open URLs in $BROWSER on mouse click")
Sheet.addCommand("", "open-url", "vd.launchBrowser(sheet.cursorValue)", "open current cursor value in $BROWSER")



'''
# Matrix Chat Sheet
A list of messages from [:underline]{sheet.sourcename}[/].

- `Enter` to push a sheet with all the messages from [:underline]{sheet.cursorRow.room.display_name}[/].
- `a` to send a message to [:underline]{sheet.cursorRow.room.display_name}[/].
'''

from visidata import vd, VisiData, Sheet, Column, ItemColumn, date, asyncthread, AttrDict, vlen, Path


vd.option('matrix_token', '', 'matrix API token')
vd.option('matrix_user_id', '', 'matrix user ID associated with token')
vd.option('matrix_device_id', 'VisiData', 'device ID associated with matrix login')

vd.matrix_client = None


@VisiData.api
def openhttp_matrix(vd, p):
    vd.importExternal('matrix_client')

    if not vd.options.matrix_token:
        from matrix_client.client import MatrixClient

        username = vd.input(f'{p.given} username: ', record=False)
        password = vd.input('password: ', record=False, display=False)

        vd.matrix_client = MatrixClient(p.given)
        matrix_token = vd.matrix_client.login(username, password, device_id=vd.options.matrix_device_id)

        vd.setPersistentOptions(matrix_user_id=username, matrix_token=matrix_token)

    vd.timeouts_before_idle = -1
    return MatrixSheet(p.base_stem, source=p)

vd.open_matrix = vd.openhttp_matrix


class MatrixRoomsSheet(Sheet):
    def iterload(self):
        yield from vd.matrix_client.get_rooms().values()


class MatrixSheet(Sheet):
    help = __doc__
    columns = [
        Column('room', width=12, getter=lambda c,r: r.room and r.room.display_name),
        ItemColumn('sender', width=0),
        Column('sender', width=10, getter=lambda c,r: r.sender.split(':')[0][1:]),
        Column('timestamp', width=16, type=date, getter=lambda c,r: r and r.origin_server_ts/1000, fmtstr='%Y-%m-%d %H:%m'),
        ItemColumn('type', width=15),
        ItemColumn('content', width=0),
        ItemColumn('content.body', width=50),
        ItemColumn('received', type=vlen, width=0),
        ItemColumn('content.msgtype'),
    ]

    @property
    def sourcename(self):
        from matrix_client.room import Room
        if isinstance(self.source, Room):
            return self.source.display_name or self.source.room_id
        else:
            return str(self.source)

    @asyncthread
    def reload(self):
        from matrix_client.client import MatrixClient
        from matrix_client.room import Room
        if not vd.matrix_client:
            vd.matrix_client = MatrixClient(self.source.given,
                                        token=self.options.matrix_token,
                                        user_id=self.options.matrix_user_id)

        if isinstance(self.source, Room):
            self.add_room(self.source)
            self.get_room_messages(self.source)
            return

        for room in vd.matrix_client.get_rooms().values():
            self.add_room(room)
            room.backfill_previous_messages(limit=1)

        vd.matrix_client.add_listener(self.global_event)
        vd.matrix_client.add_ephemeral_listener(self.global_event)

        vd.matrix_client.start_listener_thread(exception_handler=vd.exceptionCaught)

        vd.matrix_client._sync()

        vd.matrix_client.listen_for_events() # vd.matrix_client.sync(full_state=True)

    def add_room(self, room):
        room.add_listener(self.room_event)
        room.add_ephemeral_listener(self.room_event)
        room.add_state_listener(self.global_event)

    @asyncthread
    def get_room_messages(self, room):
            while room.prev_batch:
                ret = vd.matrix_client.api.get_room_messages(room.room_id, room.prev_batch, direction='b', limit=100)
                for r in ret['chunk']:
                    r['room'] = room
                    self.addRow(r)

                if 'end' not in ret or ret['end'] == room.prev_batch:
                    break
                room.prev_batch = ret['end']

    def addRow(self, r, **kwargs):
        r = AttrDict(r)
        if r.event_id not in self.event_index:
            super().addRow(r, **kwargs)
            self.event_index[r.event_id] = r

    def global_event(self, chunk):
        self.addRow(chunk)

    def room_event(self, room, chunk):
        ev = AttrDict(chunk)
        t = chunk['type']
        if t == 'm.receipt':
            for msgid, content in chunk['content'].items():
                msg = self.event_index.setdefault(msgid, {})
                if 'received' not in msg:
                    msg['received'] = {}

                for t, c in content.items():
                    assert t == 'm.read'
                    for userid, v in c.items():
                        msg['received'][(t, userid)] = v['ts']/1000
            return

        chunk['room'] = room
        self.addRow(chunk)

    def add_message(self, text):
        vd.matrix_client.send_text(text)

    def openRow(self, row):
        return MatrixSheet(row.room.display_name, source=row.room, last_id=row.event_id)


MatrixSheet.init('event_index', dict)

MatrixSheet.addCommand('a', 'add-row', 'cursorRow.room.send_text(input(cursorRow.room.display_name+"> "))', 'send chat message to current room')
vd.addMenuItems('Row > Add > send matrix message > add-msg')



from visidata import vd, Column, VisiData, ItemColumn, Path, AttrDict, date

from .gitsheet import GitSheet


@VisiData.api
def git_blame(vd, gitpath, args, **kwargs):
    if args and not args[-1].startswith('-'):
        fn = args[-1]
        return GitBlame('blame', fn, source=Path(fn), **kwargs)


class FormatColumn(Column):
    def calcValue(self, row):
        return self.expr.format(**row)


# rowdef: (hdr, orig_linenum, linenum, line)
#   hdr = { 'sha': .., 'orig_linenum': .., 'final_linenum': .. }
# source = GitSheet; .gitfile=GitFile
class GitBlame(GitSheet):
    rowtype = 'lines'
    guide = '''
        # git blame
    '''
    columns = [
        ItemColumn('sha', width=0),
        ItemColumn('orig_linenum', width=0, type=int),
        ItemColumn('final_linenum', width=0, type=int),
        ItemColumn('author', width=15),
        ItemColumn('author_time', width=13, type=date),
        FormatColumn('committer', width=0, expr='{committer} {committer_mail}'),
        ItemColumn('committer_time', width=0, type=date),
        ItemColumn('linenum', width=6, type=int),
        ItemColumn('line', width=72),
    ]

    def iterload(self):
        lines = list(self.git_lines('blame', '--porcelain', str(self.source)))
        i = 0
        headers = {}  # [sha1] -> hdr
        while i < len(lines):
            # header
            parts = lines[i].split()
            sha, orig, final = parts[:3]
            if len(parts) > 3:
                nlines_this_group = parts[3]

            if sha not in headers:
                hdr = AttrDict(sha=sha, orig_linenum=orig, final_linenum=final)
                headers[sha] = hdr
            else:
                hdr = headers[sha]

            while lines[i][0] != '\t':
                try:
                    k, v = lines[i].split(maxsplit=1)
                    k = k.replace('-', '_')
                    if '_time' in k:
                        v = int(v)
                    hdr[k] = v
                except Exception:
                    vd.status(lines[i])
                i += 1

            yield AttrDict(
                linenum=final,
                line=lines[i][1:],
                **hdr
            )
            i += 1


#GitBlame.addCommand(ENTER, 'diff-line', 'openDiff(str(gitfile), cursorRow[0]["sha"]+"^", cursorRow[0]["sha"])', 'open diff of the commit when this line changed')

#GitStatus.addCommand(None, 'git-blame', 'vd.push(GitBlame(cursorRow, source=sheet))', 'push blame for this file')



from .gitsheet import vd, GitSheet


GitSheet.options.disp_status_fmt = '{sheet.progressStatus}‹{sheet.branchStatus}› {sheet.name}| '

@GitSheet.property
def progressStatus(sheet):
    inp = sheet.gitInProgress()
    return ('[%s] ' % inp) if inp else ''


@GitSheet.property
def branchStatus(sheet):
    if hasattr(sheet.gitRootSheet, 'branch'):
        return '%s%s' % (sheet.rootSheet.branch, sheet.rootSheet.remotediff)
    return ''


@GitSheet.api
def gitInProgress(sheet):
    p = sheet.gitPath
    if not p:
        return 'no repo'
    if (p/'rebase-merge').exists() or (p/'rebase-apply/rebasing').exists():
        return 'rebasing'
    elif p/'rebase-apply'.exists():
        return 'applying'
    elif p/'CHERRY_PICK_HEAD'.exists():
        return 'cherry-picking'
    elif p/'MERGE_HEAD'.exists():
        return 'merging'
    elif p/'BISECT_LOG'.exists():
        return 'bisecting'
    return ''



from visidata import vd, VisiData, Path, ColumnItem, ESC

from .gitsheet import GitSheet


@VisiData.api
def git_grep(vd, p, args):
    return GitGrep(args[0], regex=args[0], source=p)


class GitGrep(GitSheet):
    rowtype = 'results' # rowdef: list(file, line, line_contents)
    guide = '''
        # vgit grep
        Each row on this sheet is a line matching the regex pattern `{sheet.regex}` in the tracked files of the current directory.

        - `Ctrl+O` to open _{sheet.cursorRow[0]}:{sheet.cursorRow[1]}_ in the system editor; saved changes will be reflected automatically.
    '''
    columns = [
        ColumnItem('file', 0, help='filename of the match'),
        ColumnItem('line', 1, help='line number within file'),
        ColumnItem('text', 2, width=120, help='matching line of text'),
    ]
    nKeys = 2

    def iterload(self):
        tmp = (self.topRowIndex, self.cursorRowIndex)
        for line in self.git_lines('grep', '--no-color', '-z', '--line-number', '--ignore-case', self.regex):
#            line = line.replace(ESC+'[1;31m', '[:green]')
#            line = line.replace(ESC+'[m', '[/]')
            yield list(line.split('\0'))
        self.topRowIndex, self.cursorRowIndex = tmp


GitSheet.addCommand('g/', 'git-grep', 'rex=inputRegex("git grep: "); vd.push(GitGrep(rex, regex=rex, source=sheet))', 'find in all files in this repo')
GitGrep.addCommand('Ctrl+O', 'sysopen-row', 'launchExternalEditorPath(Path(cursorRow[0]), linenum=cursorRow[1]); reload()', 'open this file in $EDITOR')
GitGrep.bindkey('Enter', 'sysopen-row')



'''
Add plot-column-ext and plot-numerics-ext to Sheet, and plot-ext to GraphSheet, to open current graph in new matplotlib window.
'''

from visidata import vd, VisiData, Sheet, GraphSheet, Progress, asyncthread

@VisiData.api
@asyncthread
def plot_seaborn(vd, rows, xcols, ycols):
    vd.status(f'plotting {len(rows)} rows using matplotlib')
    import multiprocessing
    mp = multiprocessing.Process(target=ext_plot_seaborn, args=(rows, xcols, ycols))
    mp.start()


def ext_plot_seaborn(rows, xcols, ycols):
    pd = vd.importExternal('pandas')
    plt = vd.importExternal('matplotlib.pyplot', 'matplotlib')
    sns = vd.importExternal('seaborn')

    # Set the default theme
    sns.set()

    plt.figure(figsize=(10, 6))
    plt.title('')
    plt.xticks(rotation=15)

    nerrors = 0
    nplotted = 0

    x_array = []
    y_array = []
    cat_array = []

    catcols = [c for c in xcols if not vd.isNumeric(c)]
    numcols = vd.numericCols(xcols)
    for rownum, row in enumerate(Progress(rows, 'plotting')):
        for ycol in ycols:
            try:
                if catcols:
                    k = tuple(c.getValue(row) for c in catcols)
                    if len(catcols) == 1:
                        k = k[0]
                else:
                    k = ycol.name

                graph_x = numcols[0].type(numcols[0].getValue(row)) if numcols else rownum
                graph_y = ycol.type(ycol.getValue(row))

                x_array.append(graph_x)
                y_array.append(graph_y)
                cat_array.append(k)

                nplotted += 1
            except Exception:
                nerrors += 1
                if vd.options.debug:
                    raise

    sns.scatterplot(
        x=x_array,
        y=y_array,
        hue=cat_array,
#        hue_order=df.tag.value_counts().iloc[:top].index,
#        data=tmpdf,
        s=5,
        linewidth=0,
        ).legend().set_title = (None)

    plt.show()


Sheet.addCommand('', 'plot-column-ext', 'plot_seaborn(rows, keyCols, numericCols([cursorCol]))', 'plot current numeric column on y-axis vs key columns on x-axis using matplotlib/seaborn')
Sheet.addCommand('', 'plot-numerics-ext', 'plot_seaborn(rows, keyCols, numericCols(nonKeyVisibleCols))', 'plot a graph of all visible numeric columns using matplotlib/seaborn')
GraphSheet.addCommand('', 'plot-ext', 'plot_seaborn(sourceRows, xcols, ycols)', 'replot current graph using matplotlib/seaborn')

vd.addMenuItem('Plot', 'Graph', 'using matplotlib', 'current column', 'plot-column-ext')
vd.addMenuItem('Plot', 'Graph', 'using matplotlib', 'all numeric columns', 'plot-numerics-ext')
vd.addMenuItem('Plot', 'Graph', 'replot using matplotlib', 'plot-ext')



from copy import copy
from functools import wraps

from visidata.cmdlog import CommandLog, CommandLogJsonl
from visidata import vd, UNLOADED, asyncthread, vlen
from visidata import IndexSheet, VisiData, Sheet, Path, VisiDataMetaSheet, Column, ItemColumn, AttrColumn, BaseSheet

vd.macroMode = None  # CommandLog
vd.macrobindings = {}


vd.macros = vd.StoredList(name='macros')


class MacroSheet(IndexSheet):
    guide= '''
        # Macros Sheet
        This is a list of user-defined macros.

        - `Enter` to open the current macro.
        - `d` to mark macro for delete; `z Ctrl+S` to commit.
    '''
    columns = [
        AttrColumn('binding'),
        Column('num_commands', type=vlen, width=0),
        AttrColumn('source'),
    ]
    rowtype = 'macros'  # rowdef: CommandLogJsonl
    defer = True
    nKeys = 1

    def iterload(self):
        yield from vd.macrobindings.values()

    def commitDeleteRow(self, row):
        del vd.macrobindings[row.binding]
        vd.callNoExceptions(Path(row.source).unlink)

    @asyncthread
    def putChanges(self):
        self.commitDeletes()  #1569  apply deletes early for saveSheets below

        vd.sync(vd.saveSheets(self.source, self, confirm_overwrite=False))
        self._deferredDels.clear()
        self.reload()

    def newRow(self):
        vd.fail('add macros with `m` instead')


@VisiData.lazy_property
def macrosheet(vd):
    return MacroSheet('user_macros', source=vd.macros.path)


@VisiData.api
def loadMacro(vd, p:Path):
    if p.exists():
        if p.ext == 'vd':
            vs = CommandLog(p.base_stem, source=p)
            vs.ensureLoaded()
            return vs
        elif p.ext == 'vdj':
            vs = CommandLogJsonl(p.base_stem, source=p)
            vs.ensureLoaded()
            return vs

    vd.warning(f'failed to load macro {p}')


@VisiData.api
def runMacro(vd, binding:str):
    mm = vd.macroMode
    vd.macroMode = None
    vd.replay_sync(vd.macrobindings[binding])
    vd.macroMode = mm


@VisiData.api
def setMacro(vd, ks:str, vs):
    'Set *ks* which is either a keystroke or a longname to run the cmdlog in *vs*.'
    vs.binding = ks
    vd.macrobindings[ks] = vs
    if vd.isLongname(ks):
        BaseSheet.addCommand('', ks, f'runMacro("{ks}")')
    else:
        BaseSheet.addCommand(ks, f'exec-{vs.name}', f'runMacro("{ks}")')


@CommandLogJsonl.api
def saveMacro(self, rows, ks):
        vs = copy(self)
        vs.rows = rows
        macropath = Path(vd.fnSuffix(str(Path(vd.options.visidata_dir)/ks)))
        vd.save_vdj(macropath, vs)
        vd.status(f'{ks} saved to {macropath}')
        vd.setMacro(ks, vs)
        vd.macros.append(dict(binding=ks, source=str(macropath)))
        vd.reloadMacros()
        vd.macrosheet.reload()


# needs to happen before, because the original afterexecsheet resets vd.activecommand to None
@CommandLogJsonl.before
def afterExecSheet(cmdlog, sheet, escaped, err):
    if not vd.macroMode: return
    if not vd.activeCommand: return
    if vd.activeCommand.longname == 'macro-record': return

    if vd.activeCommand.replayable:
        cmd = copy(vd.activeCommand)
        cmd.sheet = ''
        vd.macroMode.addRow(cmd)


@CommandLogJsonl.api
def startMacro(cmdlog):
    if not Path(vd.options.visidata_dir).is_dir():
        vd.fail(f'create {vd.options.visidata_dir} to save macros')
    if vd.macroMode:
        try:
            ks = vd.input('bind macro to: ', help=f'''
                # Finish recording macro
                Type in either a longname like `happy-time` (with at least one hyphen),
                   or spell out a keybinding (like `Alt+b`) manually.

                - Prefixes allowed with a keybinding: `{'  '.join(vd.allPrefixes)}`
                - Press `Ctrl+N` and then press another keystroke to spell that keystroke.
                - Press `Ctrl+C` to cancel the macro recording.
            ''')
            while ks in vd.macrobindings:
                ks = vd.input(f'{ks} already in use; set macro to keybinding: ')
            vd.cmdlog.saveMacro(vd.macroMode.rows, ks)
        finally:
            vd.macroMode = None
    else:
        vd.status("recording macro; stop recording with `m`")
        vd.macroMode = CommandLogJsonl('current_macro', rows=[])


@VisiData.before
@asyncthread
def run(vd, *args, **kwargs):
    vd.reloadMacros()


@VisiData.api
def reloadMacros(vd):
    vd.macros.reload()
    for r in vd.macros:
        vs = vd.loadMacro(Path(r.source))
        if vs:
            vd.setMacro(r.binding, vs)


Sheet.addCommand('m', 'macro-record', 'vd.cmdlog.startMacro()', 'start/stop macro recording', replay=False)
Sheet.addCommand('gm', 'macro-sheet', 'vd.push(vd.macrosheet)', 'open an index of existing macros')

vd.addMenuItems('''
    System > Record macro > macro-record
    System > Macros sheet > macro-sheet
''')



__author__ = 'Andy Craig, andycraig (https://github.com/andycraig)'


import re

from visidata import vd, Sheet, Column, floatsi, currency, date

date_fmtstrs = [
    '%Y',
    '%Y-%m',
#    '%Y-W%U',
    '%Y-%m-%d',
    '%Y-%m-%d %H',
    '%Y-%m-%d %H:%M',
    '%Y-%m-%d %H:%M:%S',
    '%Y-%m-%d %H:%M:%S.%f',
]

@Column.api
def setcol_precision(col, amount:int):
    if col.type is date:
        try:
            i = date_fmtstrs.index(col.fmtstr)
        except ValueError:
            i = 2
        col.fmtstr = date_fmtstrs[(i+amount)%len(date_fmtstrs)]
    elif col.type in (float, floatsi, currency):
        if col.fmtstr == '':
            col.fmtstr = f'%.{2 + amount}f'
        else:
            precision_str = re.match(r'%.([0-9]+)f', col.fmtstr)
            if not precision_str is None:
                col.fmtstr = f'%.{max(0, int(precision_str[1]) + amount)}f'
    else:
        vd.fail('column type must be numeric or date')


vd.addMenuItems('''
    Column > Set precision > more > setcol-precision-more
    Column > Set precision > less > setcol-precision-less
''')

Sheet.addCommand('Alt+-', 'setcol-precision-less', 'cursorCol.setcol_precision(-1)', 'show less precision in current column')
Sheet.addCommand('Alt++', 'setcol-precision-more', 'cursorCol.setcol_precision(1)', 'show more precision in current column')



'''
# TODO:
- slide-cells-left
- slide-cells-<dir>-n
- slide-cells-selected-<dir>-n
- rename "slide" to "shift"?
'''

from visidata import vd, TableSheet


@TableSheet.api
def slide_cells_right(sheet, row, vcolidx):
    for oldcol, newcol in reversed(list(zip(sheet.visibleCols[vcolidx:], sheet.visibleCols[vcolidx+1:]))):
        newcol.setValue(row, oldcol.getValue(row))

    sheet.visibleCols[vcolidx].setValue(row, None)


TableSheet.addCommand('', 'slide-cells-right', 'slide_cells_right(cursorRow, cursorVisibleColIndex)', '''
Shift individual values in current row one visible column to the right, with leftmost cell set to null.
''')

vd.addMenuItems('''
    Edit > Slide > cells > right > slide-cells-right
''')
